import sys, os, time
import numpy as np
from .mytools.MinTree import MinTree

from .._model import DMmodel
from ...util.basicutil import param_default
from ...backend import STensor

def del_block(M, rowSet ,colSet):
    M = M.tolil()

    (rs, cs) = M.nonzero()
    for i in range(len(rs)):
        if rs[i] in rowSet or cs[i] in colSet:
            M[rs[i], cs[i]] = 0
    return M.tolil()


class FlowScope( DMmodel ):
    '''Anomaly detection base on contrastively dense subgraphs, considering
    topological, temporal, and categorical (e.g. rating scores) signals, or
    any supported combinations.

    Parameters
    ----------
    graph: Graph
        Graph instance contains adjency matrix, and possible multiple signals.
    '''
    def __init__(self, graphList: list):
        self.graphnum = len(graphList)
        self.graphlist = graphList
                    

    def __str__(self):
        return str(vars(self))

    
    def run(self,  k=3, level=0, alpha=4, maxsize=-1):
        
        print("you are running with ", self.graphnum+1," partite graph")
        self.nres = []
        self.level = level
        self.alpha = alpha
        self.k = k
        self.maxsize = maxsize
        
        isCorrectInput = self.checkinput()
        if not isCorrectInput:
            print('wrong input: exit')
        else:
            self.initData()
            for i in range(self.k):
                if self.level == 0:
                    (finalsets, score) = self.fastGreedyDecreasing()
                else:
                    return print("No such level know")
    
                self.nres.append([finalsets, score])
    
                for j in range(len(self.mcurlist)):
                    self.mcurlist[j] = del_block(self.mcurlist[j], finalsets[j], finalsets[j+1])
                    self.mtranslist[j] = del_block(self.mtranslist[j], finalsets[j+1], finalsets[j])
        return self.nres
    
    
    def checkinput(self):
        isCorrectInput = True
        if not isinstance(self.maxsize, int) and not isinstance(self.maxsize,tuple):
            print('block size limit must be an integer or a tuple')
            isCorrectInput = False
        elif isinstance(self.maxsize, int):
            if self.maxsize >= 0 and self.maxsize < self.graphnum + 1:
                print('block size must be larger than the number of dimensions of multipartite graph')
                isCorrectInput = False
            if self.maxsize < 0 and self.maxsize != -1:
                print('negatives except -1 are wrong input (-1 means no size limit)')
                isCorrectInput = False
        elif isinstance(self.maxsize, tuple):
            if len(self.maxsize) != self.graphnum + 1:
                print('please enter the node size limit for each dimension of multipartite graph')
                isCorrectInput = False
            else:
                for mode in range(len(self.maxsize)):
                    if not isinstance(self.maxsize[mode], int):
                        print('The node size limit of each dimension must be an integer(-1 for no limit)')
                        isCorrectInput = False
                    if self.maxsize[mode] <= 0 and self.maxsize[mode] != -1:
                        print('node size of any dimension should be -1 or any positive integer')
                        isCorrectInput = False
        return isCorrectInput

    def initData(self):
        self.mcurlist = []
        self.mtranslist = []
        for i in range(len(self.graphlist)):
            self.mcurlist.append(self.graphlist[i].graph_tensor._data.copy().tocsr().tolil().astype(np.float64))
            self.mtranslist.append(self.graphlist[i].graph_tensor._data.copy().tocsr().tolil().transpose()) 
    
    
    def initGreedy(self):
        self.sets = []
        self.dtrees = []
        self.deltaslist = []
        self.curScorelist = []
        self.curAveScorelist = []

        self.sets.append(set(range(self.mcurlist[0].shape[0])))
        for i in range(len(self.mcurlist)-1):
            self.sets.append(set(range(self.mcurlist[i].shape[1])))
        self.sets.append(set(range(self.mcurlist[-1].shape[1])))

        s = 0
        for i in range(len(self.sets)):
            s += len(self.sets[i])
        
        rowDeltas = np.squeeze(self.mcurlist[0].sum(axis=1, dtype=np.float64).A)  # sum of A
        self.dtrees.append(MinTree(rowDeltas))
        for i in range(len(self.mcurlist)-1):
            midDeltas1 = np.squeeze(self.mcurlist[i].sum(axis=0, dtype=np.float64).A)
            midDeltas2 = np.squeeze(self.mcurlist[i+1].sum(axis=1, dtype=np.float64).A)
            self.deltaslist.append(midDeltas1)
            self.deltaslist.append(midDeltas2)
            
            mid_min = []
            mid_max = []
            for (m1, m2) in zip(midDeltas1, midDeltas2):
                temp = min(m1, m2)  # fi
                temp2 = max(m1, m2)  # qi
                mid_min.append(temp)
                mid_max.append(temp2)

            mid_min = np.array(mid_min)
            mid_max = np.array(mid_max)
            mid_priority = (1 + self.alpha) * mid_min - self.alpha * mid_max
            self.dtrees.append(MinTree(mid_priority))

            curScore1 = sum(mid_min)
            curScore2 = sum(abs(midDeltas1 - midDeltas2))
            self.curScorelist.append(curScore1)
            self.curScorelist.append(curScore2)
            curAveScore1 = curScore1 / s
            curAveScore2 = curScore2 / s
            self.curAveScorelist.append(curAveScore1)
            self.curAveScorelist.append(curAveScore2)

        colDeltas = np.squeeze(self.mcurlist[-1].sum(axis=0, dtype=np.float64).A)  # sum of C
        self.dtrees.append(MinTree(colDeltas))
        
        

    def updateConnNode(self, mold, index, mindelta):
        if mold == 0:
            # update  the  weight of connected nodes
            for j in self.mcurlist[0].rows[mindelta]:  # A
                
                if self.deltaslist[0][j] == -1:
                    continue
                new_md1 = self.deltaslist[0][j] - self.mcurlist[0][mindelta, j]
                tempmin = min(self.deltaslist[0][j], self.deltaslist[1][j])
                if (new_md1) < tempmin:  # if in-degree know is the smallist one, update the curScore1
                    self.curScorelist[0] -= (tempmin - new_md1)

                self.curScorelist[1] += abs(new_md1 - self.deltaslist[1][j]) - abs(self.deltaslist[0][j] - self.deltaslist[1][j])
                self.deltaslist[0][j] = new_md1

                # update mid_tree
                mid_delta_value = abs(self.deltaslist[0][j] - self.deltaslist[1][j])
                new_mid_w = min(self.deltaslist[0][j], self.deltaslist[1][j]) - self.alpha * mid_delta_value
                self.dtrees[1].setVal(j, new_mid_w)

            self.sets[0] -= {mindelta}  # update rowSet
            self.dtrees[0].changeVal(mindelta, float('inf'))
            self.deleted.append((index, mindelta)) 
            self.numDeleted += 1

        elif mold == 2:   # C
            # update mD1, mD2, and mid_tree
            for i in self.mtranslist[-1].rows[mindelta]:
                if self.deltaslist[-1][i] == -1:
                    continue

                new_md2 = self.deltaslist[-1][i] - self.mtranslist[-1][mindelta, i]
                tempmin = min(self.deltaslist[-1][i], self.deltaslist[-2][i])
                if (new_md2) < tempmin:
                    self.curScorelist[-2] -= (tempmin - new_md2)

                self.curScorelist[-1] += abs(new_md2 - self.deltaslist[-2][i]) - abs(self.deltaslist[-1][i] - self.deltaslist[-2][i])
                self.deltaslist[-1][i] = new_md2

                mid_delta_value = abs(self.deltaslist[-1][i] - self.deltaslist[-2][i])
                new_mid_w = min(self.deltaslist[-1][i], self.deltaslist[-2][i]) - self.alpha * mid_delta_value
                self.dtrees[-2].setVal(i, new_mid_w)

            self.sets[-1] -= {mindelta}
            self.dtrees[-1].changeVal(mindelta, float('inf'))
            self.deleted.append((index, mindelta))
            self.numDeleted += 1

        elif mold == 1:

            self.curScorelist[2 * index - 2] -= min(self.deltaslist[2 * index - 2][mindelta],
                                                    self.deltaslist[2 * index - 1][mindelta])
            self.curScorelist[2 * index - 1] -= abs(
                self.deltaslist[2 * index - 2][mindelta] - self.deltaslist[2 * index - 1][mindelta])

            self.deltaslist[2 * index - 2][mindelta] = -1
            self.deltaslist[2 * index - 1][mindelta] = -1

            for j in self.mcurlist[index].rows[mindelta]:
                if index < (len(self.dtrees) - 2):
                    if self.deltaslist[2 * index][j] == -1:
                        continue
                    new_md1 = self.deltaslist[2 * index][j] - self.mcurlist[index][mindelta, j]
                    tempmin = min(self.deltaslist[2 * index][j], self.deltaslist[2 * index + 1][j])
                    if (new_md1) < tempmin:  # if in-degree know is the smallist one, update the curScore1
                        self.curScorelist[2 * index] -= (tempmin - new_md1)

                    self.curScorelist[2 * index + 1] += abs(new_md1 - self.deltaslist[2 * index + 1][j]) - abs(
                        self.deltaslist[2 * index][j] - self.deltaslist[2 * index + 1][j])
                    self.deltaslist[2 * index][j] = new_md1


                    mid_delta_value = abs(self.deltaslist[2 * index][j] - self.deltaslist[2 * index + 1][j])
                    new_mid_w = min(self.deltaslist[2 * index][j], self.deltaslist[2 * index + 1][j]) - self.alpha * mid_delta_value
                    self.dtrees[index + 1].setVal(j, new_mid_w)
                else:
                    self.dtrees[index + 1].changeVal(j, -self.mcurlist[index][mindelta, j])

            for j in self.mtranslist[index - 1].rows[mindelta]:
                if index > 1:
                    if self.deltaslist[2 * index - 3][j] == -1:
                        continue

                    new_md2 = self.deltaslist[2 * index - 3][j] - self.mtranslist[index - 1][mindelta, j]
                    tempmin = min(self.deltaslist[2 * index - 3][j], self.deltaslist[2 * index - 4][j])
                    if (new_md2) < tempmin:
                        self.curScorelist[2 * index - 4] -= (tempmin - new_md2)

                    self.curScorelist[2 * index - 3] += abs(new_md2 - self.deltaslist[2 * index - 4][j]) - abs(
                        self.deltaslist[2 * index - 3][j] - self.deltaslist[2 * index - 4][j])
                    self.deltaslist[2 * index - 3][j] = new_md2

                    mid_delta_value = abs(self.deltaslist[2 * index - 3][j] - self.deltaslist[2 * index - 4][j])
                    new_mid_w = min(self.deltaslist[2 * index - 3][j], self.deltaslist[2 * index - 4][j]) - self.alpha * mid_delta_value
                    self.dtrees[index - 1].setVal(j, new_mid_w)
                else:
                    self.dtrees[index - 1].changeVal(j, -self.mtranslist[index - 1][mindelta, j])

            self.sets[index] -= {mindelta}
            self.dtrees[index].changeVal(mindelta, float('inf'))
            self.deleted.append((index, mindelta))
            self.numDeleted += 1
            
    
    # find the smallest one in all set
    def findmin(self):
        resmin = []
        for i in range(len(self.dtrees)):
            (next, delt) = self.dtrees[i].getMin()
            resmin.append((next, delt))

        min1 = resmin[0][1] * (1 + self.alpha) 
        min2 = resmin[-1][1] * (1 + self.alpha)
        min_weight = min1
        minidx = resmin[0][0]
        idx = 0
        mold = 0
        
        if min2 < min_weight:
            min_weight = min2
            idx = len(resmin) - 1
            mold = 2
            minidx = resmin[-1][0]
        for i in range(len(resmin) - 2):
            if resmin[i + 1][1] < min_weight:
                min_weight = resmin[i + 1][1]
                minidx = resmin[i + 1][0]
                idx = i + 1
                mold = 1
        
        return mold, idx, minidx
        
        
    def checkset(self):
        res = True
        for i in range(len(self.sets)):
            res = self.sets[i] and res
        return res


    def checksize(self):
        isCorrectSize = True
        if isinstance(self.maxsize, tuple):
            for mode in range(len(self.maxsize)):
                if self.maxsize[mode] != -1 and len(self.sets[mode]) > self.maxsize[mode]:
                    isCorrectSize = False
                    break
        if isinstance(self.maxsize, int) and self.maxsize > 0:
            nodesum = sum([len(self.sets[mode]) for mode in range(len(self.sets))])
            if nodesum > self.maxsize:
                isCorrectSize = False
        return isCorrectSize
    
    # use to print the result of algorithm
    def printres(self):
        print('best delete number : ', self.bestNumDeleted)
        res1 = 'nodes number remaining:  '
        for i in range(len(self.sets)):
            res1 += str(len(self.sets[i])) + ' '
        print(res1)
        res2 = 'matrix mass remaining:  '
        for i in range(len(self.curScorelist)):
            res2 += str((self.curScorelist[i])) + ' '
        print(res2)

        print('best score of g(S): ', self.bestAveScore)
        res3 = 'min value of the tree :  '
        for i in range(len(self.dtrees)):
            res3 += str((self.dtrees[i].getMin())) + ' '
        print(res3)
  
    
        
    def fastGreedyDecreasing(self):
        print('this is the cpu version of FlowScope')  
        print('start  greedy')

        self.initGreedy()

        self.bestAveScore = float('-inf')

        self.numDeleted = 0
        self.deleted = []
        self.bestNumDeleted = 0

        curall = 0
        for i in range(0, len(self.curScorelist), 2):
            curall += self.curAveScorelist[i] - self.alpha * self.curAveScorelist[i+1]
        print('initial score of g(S):', curall)
        while self.checkset():  # repeat deleting until one node set in null

            mold, idx, minidx = self.findmin()

            self.updateConnNode(mold=mold, index=idx, mindelta=minidx)
            
            s = 0
            for i in range(len(self.sets)):
                s += len(self.sets[i])
            if (s) > 0:
                for i in range(0, len(self.curAveScorelist), 2):
                    self.curAveScorelist[i] = self.curScorelist[i] / s
                    self.curAveScorelist[i+1] = self.curScorelist[i+1] / s
            else:
                print("something wrong in FlowScope")
                for i in range(0, len(self.curAveScorelist), 2):
                    self.curAveScorelist[i] = 0
                    self.curAveScorelist[i+1] = 0
            isCorrectSize = self.checksize()
            if isCorrectSize:
                curAveScore = 0
                for i in range(0, len(self.curAveScorelist), 2):
                    curAveScore += self.curAveScorelist[i] - self.alpha * self.curAveScorelist[i + 1]
                if curAveScore >= self.bestAveScore:
                    self.bestNumDeleted = self.numDeleted
                    self.bestAveScore = curAveScore
                
        self.printres()

        finalsets = []
        finalsets.append(set(range(self.mcurlist[0].shape[0])))
        for i in range(len(self.mcurlist) - 1):
            finalsets.append(set(range(self.mcurlist[i].shape[1])))
        finalsets.append(set(range(self.mcurlist[-1].shape[1])))
        
        for i in range(self.bestNumDeleted):
            finalsets[self.deleted[i][0]].remove(self.deleted[i][1])
        
        res = 'the block size is: '
        for i in range(len(finalsets)):
            res += str(len(finalsets[i])) + ' '
        print(res+'\n')
        
        return finalsets, self.bestAveScore


    def anomaly_detection(self, k=3, level=0, alpha=4, maxsize=-1):
        return self.run(k, level, alpha, maxsize)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   train.py
@Desc    :   Implementation for training task.
'''

# here put the import lib

from . import MODEL_PATH

from ._task import Task
from enum import Enum


class Train(Task):
    '''Implementation for training task.
    '''

    def run(self, *args, **kwargs):
        '''Call train function of selected model.

        If not implemented, raise an exception by calling parent run.
        '''
        if "train" in dir(self.model):
            return self.model.train(*args, **kwargs)
        else:
            return super().run("train")

    @classmethod
    def list_models(cls):
        '''List models which can do training task.
        '''
        for name in TrainPolicy.__members__:
            print(name)


class TrainPolicy(Enum):
    '''Registration for path of models who can do train task.
    '''
    BeatGAN = MODEL_PATH + '.beatgan'



import numpy as np
from .mytools.MinTree import MinTree

from .._model import DMmodel
from ...util.basicutil import param_default
import copy
from collections import defaultdict
from .util import get_real_res, preprocess_data, loadtxt2res, saveres2txt, get_zero_matrix
import pdb

import sparse as sp

class CubeFlow(DMmodel):
    '''Anomaly detection base on contrastively dense subgraphs, considering
    topological, temporal, and categorical (e.g. rating scores) signals, or
    any supported combinations.

    Parameters
    ----------
    graph: Graph
        Graph instance contains adjency matrix, and possible multiple signals.
    '''
    def __init__(self, stensorList:list, **params):
        self.alpha = param_default(params, 'alpha', 0.8)
        self.k = param_default(params, 'k', 1)
        self.dim = param_default(params, 'dim', 3)
        self.outpath = param_default(params, 'outpath', './output/CFD-3/')
        self.graphlist, self.mt_dictList, self.m_mt_dict, self.m_mtSize_dict_ori = preprocess_data(stensorList, self.dim)
        self.graphnum = len(self.graphlist)
        
    def __str__(self):
        return str(vars(self))
    
    def anomaly_detection(self):
        return self.run()

    def run(self, del_type=0, maxsize=(-1,-1,-1), is_find_all_blocks=False):
        print("you are running with ", self.graphnum + 1, "partite graph")
        self.del_type = del_type
        self.nres = []
        self.size_max_limit = maxsize
        self.has_limit = False
        self.is_find_all_blocks = is_find_all_blocks
        self.m_mtSize_dict_delete = copy.deepcopy(self.m_mtSize_dict_ori)
        
        if type(self.size_max_limit) == int and self.size_max_limit != -1:
            print('Block size sum (X+Y+Z) limit:', self.size_max_limit)
            self.has_limit = True
            
        elif type(self.size_max_limit) == tuple:
            if len(self.size_max_limit) != 3:
                raise Exception('The length of size_max_limit tuple should be 3!')
            
            if not (self.size_max_limit[0] == -1 and self.size_max_limit[1] == -1 and self.size_max_limit[2] == -1):
                print(f'Block size limits: (X:{self.size_max_limit[0]}, Y:{self.size_max_limit[1]}, Z:{self.size_max_limit[2]})')
                self.has_limit = True
        else:
            print('No block size limit.')
        
        self.initData()
        if not is_find_all_blocks:
            for i in range(self.k):
                finalsets, score = self.fastGreedyDecreasing()
                real_res = get_real_res([finalsets, score], self.dim, self.mt_dictList, self.outpath)
                        
                self.nres.append(real_res)
                self.del_block(finalsets)
        else:
            if self.del_type == 0:
                raise ValueError("You cannot find all blocks with del_type equals to 0!")
            
            if self.has_limit:
                curr_i = 0
                while self.checkset(self.sets_ori):
                    finalsets, score = self.fastGreedyDecreasing()
                    real_res = get_real_res([finalsets, score], self.dim, self.mt_dictList, self.outpath)
                    
                    self.nres.append(real_res)
                    self.del_block(finalsets)
                    curr_i += 1
            else:
                print('ERROR: You cannot find all blocks without specifying the maximum size of the block!')
                return 
        return self.nres

    def initData(self):
        AM_tensor_data = self.graphlist[0].graph_tensor._data.copy()  # A, M, T, money
        CM_tensor_data = self.graphlist[1].graph_tensor._data.copy()  # C, M, T, money
        self.AM_mat = AM_tensor_data.to_scipy_sparse().tolil()  # A,(M,T),money
        self.CM_mat = CM_tensor_data.to_scipy_sparse().tolil()  # C,(M,T),money
        self.AM_tran_mat = AM_tensor_data.to_scipy_sparse().T.tolil()  # A,(M,T),money
        self.CM_tran_mat = CM_tensor_data.to_scipy_sparse().T.tolil()  # C,(M,T),money
        print(self.AM_mat.shape, self.CM_mat.shape)
        
        if self.has_limit:
            self.m_set_ori = set(list(self.m_mt_dict.keys()))
            print('m size:', len(self.m_set_ori))
        else:
            self.m_set_ori = set()
        
        self.sets_ori = []
        if self.is_find_all_blocks:
            midDeltas1, midDeltas2 = np.array(np.squeeze(self.AM_tran_mat.sum(axis=1).A)), \
                                            np.array(np.squeeze(self.CM_tran_mat.sum(axis=1).A))
            
            self.sets_ori.append(set(range(self.AM_mat.shape[0])))
            self.sets_ori.append(set(range(max(len(midDeltas1), len(midDeltas2)))))
            self.sets_ori.append(set(range(self.CM_mat.shape[0])))
            print(len(self.sets_ori[0]), len(self.sets_ori[1]), len(self.sets_ori[2]))

    # find the smallest one in all set
    def findmin(self):
        min_tree_i = -1
        min_indices = -1
        min_value = float('inf')
        for i in range(len(self.dtrees)):
            index, value = self.dtrees[i].getMin()
            if value < min_value:
                min_value = value
                min_tree_i = i
                min_indices = index
    
        return min_tree_i, min_indices

    def checkset(self, sets):
        res = True
        for i in range(len(sets)):
            if len(sets[i]) == 0:
                res = False
                break
        return res
    
    def checkset_size_max_limit(self):
        now_x_size, now_y_size, now_z_size = len(self.sets[0]), len(self.m_set), len(self.sets[2])
        
        if type(self.size_max_limit) == int:
            if (now_x_size + now_y_size + now_z_size) > self.size_max_limit:
                return False
            
        elif type(self.size_max_limit) == tuple:
            x_max, y_max, z_max = self.size_max_limit
            if x_max != -1 and now_x_size > x_max:
                return False
            if y_max != -1 and now_y_size > y_max:
                return False
            if z_max != -1 and now_z_size > z_max:
                return False
            
        return True

    def initGreedy(self):
        self.sets = []
        self.dtrees = []
        self.mid_min = []
        self.midDeltas1 = []
        self.midDeltas2 = []
        self.curScore1 = 0  # sum of f
        self.curScore2 = 0  # sum of (q-f)
        self.bestAveScore = 0
        self.bestAveSocre1 = 0 # sum of f when algorithm gets best score
        self.bestAveSocre2 = 0 # sum of (q-f) when algorithm gets best score
        self.bestNumDeleted = defaultdict(int)
        
        rowDeltas = np.squeeze(self.AM_mat.sum(axis=1).A)  # sum of A
        self.dtrees.append(MinTree(rowDeltas))  # tree 0: u
        
        self.midDeltas1, self.midDeltas2 = np.array(np.squeeze(self.AM_tran_mat.sum(axis=1).A)), \
                                           np.array(np.squeeze(self.CM_tran_mat.sum(axis=1).A))
        difflen = abs(len(self.midDeltas1) - len(self.midDeltas2))
        if len(self.midDeltas1) > len(self.midDeltas2):
            self.midDeltas2 = np.append(self.midDeltas2, np.zeros((1, difflen)))
        if len(self.midDeltas1) < len(self.midDeltas2):
            self.midDeltas1 = np.append(self.midDeltas1, np.zeros((1, difflen)))
        print(len(self.midDeltas1), len(self.midDeltas2))
        
        mid_min, mid_max = [], []
        for m1, m2 in zip(self.midDeltas1, self.midDeltas2):
            mid_min.append(min(m1, m2))  # fi
            mid_max.append(max(m1, m2))  # qi
        self.mid_min, self.mid_max = np.array(mid_min), np.array(mid_max)
        mid_priority = self.mid_min - self.alpha * self.mid_max  # f- alpha * q
        self.dtrees.append(MinTree(mid_priority))  # tree 1: (v,t)
        
        colDeltas = np.squeeze(self.CM_mat.sum(axis=1).A)  # sum of C
        self.dtrees.append(MinTree(colDeltas))  # tree 2: w

        self.curScore1 = sum(mid_min) 
        self.curScore2 = sum(abs(self.midDeltas1 - self.midDeltas2))
        
        self.sets.append(set(range(self.AM_mat.shape[0])))
        self.sets.append(set(range(len(self.mid_min))))
        self.sets.append(set(range(self.CM_mat.shape[0])))
        s = sum([len(self.sets[i]) for i in range(len(self.sets))])
        print('s size: ', s)
        
        if self.has_limit:
            self.m_mtSize_dict = copy.deepcopy(self.m_mtSize_dict_ori)
            self.m_set = set(list(self.m_mt_dict.keys()))
        
        curAveScore = float('-inf') # 看一下是否更改把numDeleted的初始值
        
        if self.has_limit:
            if self.checkset_size_max_limit():
                curAveScore = ((1 - self.alpha) * self.curScore1 - self.alpha * self.curScore2) / s
        else:
            curAveScore = ((1 - self.alpha) * self.curScore1 - self.alpha * self.curScore2) / s
            
        print('initial score of g: ', curAveScore)
        self.bestAveScore = curAveScore
        self.bestAveSocre1 = self.curScore1
        self.bestAveSocre2 = self.curScore2
        
    def updateConnNode(self, min_tree_i, idx):
        if min_tree_i == 0:  # delete a node from u
            # update  the  weight of connected nodes
            for j in self.AM_mat.rows[idx]: 
                if self.midDeltas1[j] <= 0:
                    continue
                self.curScore1 -= self.mid_min[j] 
                self.curScore2 -= abs(self.midDeltas1[j] - self.midDeltas2[j])
                
                self.midDeltas1[j] -= self.AM_mat[idx, j]
                f, q = min(self.midDeltas1[j], self.midDeltas2[j]), max(self.midDeltas1[j], self.midDeltas2[j])
                self.mid_min[j] = f
                self.curScore1 += f
                self.curScore2 += (q - f)
                
                new_mid_w = f - self.alpha * q
                self.dtrees[1].setVal(j, new_mid_w)
        elif min_tree_i == 2:  # delete a node from w
            # update mD1, mD2, and mid_tree
            for j in self.CM_mat.rows[idx]:
                if self.midDeltas2[j] <= 0:
                    continue
                self.curScore1 -= self.mid_min[j]
                self.curScore2 -= abs(self.midDeltas1[j] - self.midDeltas2[j])
                
                self.midDeltas2[j] -= self.CM_mat[idx, j]
                f, q = min(self.midDeltas1[j], self.midDeltas2[j]), max(self.midDeltas1[j], self.midDeltas2[j])
                self.mid_min[j] = f
                self.curScore1 += f
                self.curScore2 += (q - f)

                new_mid_w = f - self.alpha * q
                self.dtrees[1].setVal(j, new_mid_w)
        elif min_tree_i == 1:  # delete a node from M
            if idx < self.AM_tran_mat.shape[0]:
                for j in self.AM_tran_mat.rows[idx]:
                    self.dtrees[0].changeVal(j, -self.AM_tran_mat[idx, j])
            if idx < self.CM_tran_mat.shape[0]:
                for j in self.CM_tran_mat.rows[idx]:  # update connected nodes in C
                    self.dtrees[2].changeVal(j, -self.CM_tran_mat[idx, j])
            
            self.curScore1 -= self.mid_min[idx]
            self.curScore2 -= abs(self.midDeltas1[idx] - self.midDeltas2[idx])
            self.midDeltas1[idx] = -1
            self.midDeltas2[idx] = -1
        self.sets[min_tree_i] -= {idx}
        self.dtrees[min_tree_i].changeVal(idx, float('inf'))
        self.deleted[min_tree_i].append(idx)
        self.numDeleted[min_tree_i] += 1
        
        if self.has_limit and min_tree_i == 1: # remove corresponding m in m_set
            # map mt to m
            tmp_m = self.mt_dictList[0][idx]
            self.m_mtSize_dict[tmp_m] -= 1
            
            if self.m_mtSize_dict[tmp_m] <= 0: # all mt have been removed
                self.m_set -= {tmp_m}
    
    def fastGreedyDecreasing(self):
        print('this is the cpu version of CubeFlow')  
        print('start  greedy')

        self.initGreedy()
        self.numDeleted = defaultdict(int)
        self.deleted = {}
        for i in range(len(self.sets)):
            self.deleted[i] = []
        finalsets = copy.deepcopy(self.sets)
        
        # repeat deleting until one node set is empty
        while self.checkset(self.sets): 
            min_tree_i, idx = self.findmin()
            self.updateConnNode(min_tree_i, idx)
            s = sum([len(self.sets[i]) for i in range(len(self.sets))])
            curAveScore = ((1 - self.alpha) * self.curScore1 - self.alpha * self.curScore2) / s
                
            if self.has_limit:
                if curAveScore >= self.bestAveScore and self.checkset_size_max_limit() and self.checkset(self.sets):
                    for i in range(len(self.sets)):
                        self.bestNumDeleted[i] = self.numDeleted[i]
                    self.bestAveScore = curAveScore
                    self.bestAveSocre1 = self.curScore1
                    self.bestAveSocre2 = self.curScore2
            else:
                if curAveScore >= self.bestAveScore and self.checkset(self.sets):
                    for i in range(len(self.sets)):
                        self.bestNumDeleted[i] = self.numDeleted[i]
                    self.bestAveScore = curAveScore
                    self.bestAveSocre1 = self.curScore1
                    self.bestAveSocre2 = self.curScore2
                
        
        print('best delete number : ', self.bestNumDeleted)
        print('nodes number remaining:  ', len(self.sets[0]), len(self.sets[1]), len(self.sets[2]))
        print(f'best score of g(S): {self.bestAveScore}, f: {self.bestAveSocre1}, q-f: {self.bestAveSocre2}')
        
        for i in range(len(finalsets)):
            best_deleted = self.deleted[i][:self.bestNumDeleted[i]]
            finalsets[i] = finalsets[i] - set(best_deleted)
        print('size of found subgraph:  ', len(finalsets[0]), len(finalsets[1]), len(finalsets[2]))
        
        if sum(self.bestNumDeleted.values()) == 0:  
            # 没有找到任何子图满足限制条件，返回原始子图。例如，限制了m的大小，算法把a集合删空了之后m大小依然大于限制。
            print('Do not find any sub-block that meets the settings.')
            for i in range(len(finalsets)):
                finalsets[i] = copy.deepcopy(self.sets_ori[i])
                
            return finalsets, float('-inf')
        
        return finalsets, self.bestAveScore

    def del_block(self, finalsets):
        a_set, mt_set, c_set = set(finalsets[0]), set(finalsets[1]), set(finalsets[2])
        
        if self.del_type == 0: # del find edges
            a_mt_nnz = self.AM_mat.nonzero()
            for a, mt in zip(a_mt_nnz[0], a_mt_nnz[1]):
                if a in a_set and mt in mt_set:
                    self.AM_mat[a, mt] = 0
                    self.AM_tran_mat[mt, a] = 0
                    ##################
                    if self.has_limit:
                        tmp_m = self.mt_dictList[0][mt]
                        self.m_mtSize_dict_delete[tmp_m] -= 1
                        if self.m_mtSize_dict_delete[tmp_m] <= 0:  # all mt have been removed
                            self.m_set_ori -= {tmp_m}
                    
            c_mt_nnz = self.CM_mat.nonzero()
            for c, mt in zip(c_mt_nnz[0], c_mt_nnz[1]):
                if c in c_set and mt in mt_set:
                    self.CM_mat[c, mt] = 0
                    self.CM_tran_mat[mt, c] = 0
                    ###############
                    if self.has_limit:
                        tmp_m = self.mt_dictList[0][mt]
                        self.m_mtSize_dict_delete[tmp_m] -= 1
                        if self.m_mtSize_dict_delete[tmp_m] <= 0:  # all mt have been removed
                            self.m_set_ori -= {tmp_m}
                    
        elif self.del_type == 1: # del find points
            # map mt => m
            m_set = []
            for mt in mt_set:
                m_set.append(self.mt_dictList[0][mt])
            m_set = set(m_set)
            print('find m len:', len(m_set))
            
            # find all mt
            total_mt_set = []
            for m in m_set:
                total_mt_set.extend(self.m_mt_dict[m])
            total_mt_set = set(total_mt_set)
            
            if self.has_limit:
                self.m_set_ori -= m_set
            
            if self.is_find_all_blocks:
                self.sets_ori[0] -= a_set
                self.sets_ori[1] -= total_mt_set
                self.sets_ori[2] -= c_set
            
            a_mt_nnz = self.AM_mat.nonzero()
            for a, mt in zip(a_mt_nnz[0], a_mt_nnz[1]):
                if a in a_set or mt in total_mt_set:
                    self.AM_mat[a, mt] = 0
                    self.AM_tran_mat[mt, a] = 0

            c_mt_nnz = self.CM_mat.nonzero()
            for c, mt in zip(c_mt_nnz[0], c_mt_nnz[1]):
                if c in c_set or mt in total_mt_set:
                    self.CM_mat[c, mt] = 0
                    self.CM_tran_mat[mt, c] = 0   
                    
        elif self.del_type == 2:
            m_set = []
            for mt in mt_set:
                m_set.append(self.mt_dictList[0][mt])
            m_set = set(m_set)
            print('find m len:', len(m_set))

            # find all mt
            total_mt_set = []
            for m in m_set:
                total_mt_set.extend(self.m_mt_dict[m])
            total_mt_set = set(total_mt_set)

            if self.has_limit:
                self.m_set_ori -= m_set

            if self.is_find_all_blocks:
                self.sets_ori[0] -= a_set
                self.sets_ori[1] -= total_mt_set
                self.sets_ori[2] -= c_set
                
            a_zero_size, mt_zero_size = self.AM_mat.shape
            c_zero_size, _ = self.CM_mat.shape
            A_zero_mat = get_zero_matrix(a_zero_size, list(a_set))
            MT_zero_mat = get_zero_matrix(mt_zero_size, list(total_mt_set))
            C_zero_mat = get_zero_matrix(c_zero_size, list(c_set))
            
            self.AM_mat = A_zero_mat * self.AM_mat * MT_zero_mat
            self.AM_tran_mat = MT_zero_mat * self.AM_tran_mat * A_zero_mat
            
            self.CM_mat = C_zero_mat * self.CM_mat * MT_zero_mat
            self.CM_tran_mat = MT_zero_mat * self.CM_tran_mat * C_zero_mat
            
        else:
            raise Exception('Unknown del_type! (del_type should be in [0, 1])')

    @classmethod
    def loadtxt2res(cls, *args, **kwargs):
        return loadtxt2res(*args, **kwargs)

    @classmethod
    def saveres2txt(cls, *args, **kwargs):
        return saveres2txt(*args, **kwargs)



import sys
import numpy as np
import scipy as sci
from scipy.sparse import coo_matrix, csc_matrix
# from gendenseblock import *
from .mytools.ioutil import myreadfile
import math
from spartan.backend import STensor, DTensor

class MultiEedgePropBiGraph:
    def __init__(self, wadjm):
        self.wadjm = wadjm.tocsr().astype(np.float64) #weighted adjacent matrix
        self.nU , self.nV = wadjm.shape
        self.indegrees = self.wadjm.sum(0).getA1()
        self.inbd=2 # the objects that has at least 2 edges are considered

    #@profile
    def load_from_edgeproperty(self, profnm, mtype=coo_matrix,
                               dtype=int):
        'load the graph edge property, time stamps, ratings, or text vector'
        self.idstartzero = True #record for output recovery
        offset = -1 if self.idstartzero is False else 0
        'sparse matrix has special meaning of 0, so property index start from 1'
        self.eprop = [np.array([])] #make the idx start from 1 in sparse matrix
        with myreadfile(profnm, 'rb') as fin:
            idx=1
            x,y,data=[],[],[]
            for line in fin:
                um, prop = line.strip().split(':')
                u, m = um.split('-')
                u = int(u)+offset
                m = int(m)+offset
                x.append(u)
                y.append(m)
                data.append(idx) #data store the index of edge properties
                prop = np.array(prop.strip().split()).astype(dtype)
                self.eprop.append(prop)
                idx += 1
            fin.close()
            self.edgeidxm = mtype((data, (x,y)), shape=(max(x)+1, max(y)+1))
            self.edgeidxmr = self.edgeidxm.tocsr()
            self.edgeidxmc = self.edgeidxm.tocsc()
            self.edgeidxml = self.edgeidxm.tolil()
            self.edgeidxmlt = self.edgeidxm.transpose().tolil()
            self.eprop = np.array(self.eprop)
        pass

    def trans_array_to_edgeproperty(self, arr:STensor, mtype=coo_matrix, dtype=int):
        ''' transform the ndarray arr to a pmi matrix.
        Do not support idstartzero. Use id as it is.
        '''
        'index of (v, v, attribute), and frequency of it '
        coords, freqs = arr.coords, arr.data
        # todo: it is better to have reduce func to have all prop combine
        uvdict = {}
        eprop=[[]] # make the first start from 1
        for i in range(len(freqs)):
            u,v,p = coords[:, i]
            if (u,v) not in uvdict:
                uvdict[(u,v)] = len(uvdict) + 1 #start from 1
                eprop.append([])
            eprop[ uvdict[(u,v)] ] += [p]*freqs[i]

        x,y = np.array(list(uvdict.keys())).T
        data = list(uvdict.values())
        self.edgeidxm = mtype((data, (x,y)), shape=(max(x)+1, max(y)+1) )
        self.edgeidxmr = self.edgeidxm.tocsr()
        self.edgeidxmc = self.edgeidxm.tocsc()
        self.edgeidxml = self.edgeidxm.tolil()
        self.edgeidxmlt = self.edgeidxm.transpose().tolil()
        self.eprop = np.array(eprop)
        pass

    #@profile
    def setup_rate4all_sinks(self):
        '''set up the rating property for all sinks'''
        propvals=set() #vacabulary size or score space
        for vs in self.eprop:
            for v in set(vs):
                propvals.add(v)
        self.propvals = np.array(sorted(list(propvals)))
        'assume score is [1,5], and  arounding real scores into 3 catagories, '
        if not (max(propvals)==5 and min(propvals)>=1):
            print('Warning: rating scores are not in [1,5]. They are [{}]'.\
                    format(', '.join(map(str, propvals))))

        '(1, 1.5, 2), (2.5, 3, 3.5), (4, 4.5, 5)'
        for i in range(len(self.eprop)):
            if min(propvals)<1:
                self.eprop[i] = np.digitize(self.eprop[i], bins=[0,2.5,4,5.01])-1
            else:
                self.eprop[i] = np.digitize(self.eprop[i], bins=[1,2.5,4,5.01])-1
        allmlt = self.edgeidxmlt #all susp msg matrix
        'effect sinks'
        cols = np.argwhere(self.indegrees>=self.inbd).flatten()
        self.inbdcolset = set(cols)
        apv = {}  #all property values
        ahists={} #all histograms of sinks
        amean, avar = np.zeros(self.nV, dtype=np.float64), \
                np.zeros(self.nV, dtype=np.float64)
        for i in cols:
            aidx = allmlt.data[i]
            apvi = np.concatenate(self.eprop[aidx]) #no np.sort
            apv[i]= apvi
            amean[i] = apvi.mean()
            avar[i] = apvi.var()
            ahists[i] = np.bincount(apvi, minlength=3)
        self.amean, self.avar = amean, avar
        self.ahists = ahists
        self.apv = apv
        return

    #@profile
    def setup_ts4all_sinks(self, twait, bins='auto'):
        'calculate the one-time values for every sink, like bursting, dying, drop'
        maxts = [np.max(t) for t in self.eprop[1:]]
        self.endt = max(maxts)
        self.twait = twait
        allmlt = self.edgeidxmlt #all susp msg matrix
        'effect sinks'
        cols = np.argwhere(self.indegrees>=self.inbd).flatten()
        self.inbdcolset = set(cols)
        apv = {}
        awakburstpt, burstvals, burstslops, ainbursts={},{},{},{}
        dyingpt = {}
        dropslops, dropfalls=np.zeros(self.nV, dtype=np.float64), \
                np.zeros(self.nV, dtype=np.float64)
        amean, avar = np.zeros(self.nV, dtype=np.float64), \
                np.zeros(self.nV, dtype=np.float64)
        for i in cols:
            aidx = allmlt.data[i]
            aumts = np.concatenate(self.eprop[aidx]) #no sort
            apv[i]= aumts
            amean[i] = aumts.mean()
            avar[i] = aumts.var()
            'awaking bursting points and values, debugpt for debug'
            abpts, bvs, slops, debugpt = awakburstpoints_recur(aumts, bins=bins)
            awakburstpt[i], burstvals[i], burstslops[i] =abpts, bvs, slops
            cnts=[]
            for abpt in abpts:
                '#of edges involve in bursting'
                left, right = abpt
                cnt = ((aumts>=left) & (aumts<=right)).sum()
                cnts.append(cnt)
            ainbursts[i]=np.array(cnts)
            dropfall, dropt, slop = \
                burstmaxdying_recur(aumts, endt=self.endt, twait=self.twait, bins=bins)
            dyingpt[i] = dropt
            dropslops[i], dropfalls[i] = slop, dropfall

        self.amean, self.avar = amean, avar
        self.apv = apv
        self.awakeburstpt, self.burstvals, self.burstslops, self.ainbursts = \
                awakburstpt, burstvals, burstslops, ainbursts
        self.dyingpt, self.dropslops, self.dropfalls = \
                dyingpt, dropslops, dropfalls
        return

    'this is only called once, always put into the init/start func'
    #@profile
    def setupsuspects(self, users):
        self.suspuser = np.array(users)
        self.deltacols, self.delcols = [], set()
        if len(self.suspuser) ==0:
            self.spv = {}
            return
        #suspmlt = self.edgeidxmr[self.suspuser].transpose().tolil()
        suspmlt = self.edgeidxml[self.suspuser].transpose()
        colwsum = self.wadjm[self.suspuser].sum(0).getA1()
        cols = np.where(colwsum>= self.inbd)[0]
        cols = set(cols) & self.inbdcolset
        spv = {}
        for col in cols:
            spids = suspmlt.data[col]
            #property indices of suspect sink
            #only consider those objects have more than inbd edges with suspusers
            sumts = np.concatenate(self.eprop[spids])
            spv[col]=sumts
        self.spv = spv
        return

    'must be effecient, shared among rating, ts, text'
    #@profile
    def deltasuspects(self, z, yusers, add=True):
        self.suspuser = yusers
        zmat = self.edgeidxmr[z]
        cols = zmat.nonzero()[1]
        deltacols, delcols = [], set()
        i = -1
        for col in cols:
            i += 1
            if col not in self.inbdcolset:
                continue
            spid = zmat.data[i]
            if add:
                self.spv[col]= np.concatenate((self.spv[col],self.eprop[spid]))\
                        if col in self.spv else self.eprop[spid]
                deltacols.append(col)
            else:
                if col not in self.spv:
                    continue #donot added in the initial
                #minus
                self.spv[col] = list(self.spv[col])
                for e in self.eprop[spid]:
                    self.spv[col].remove(e)
                if len(self.spv[col])==0:
                    self.spv.pop(col, None)
                    delcols.add(col)
                else:
                    deltacols.append(col)

        self.deltacols, self.delcols = deltacols, delcols
        return

    #@profile
    def suspratedivergence(self, neutral=False, delta=False):
        '''calculate the diverse of ratings betwee A and U\A
           scaling=False
        '''
        if delta and hasattr(self, 'ratediv'):
            cols, delcols = self.deltacols, self.delcols
            ratediv = self.ratediv
            if len(self.spv) < 1:
                self.ratediv[0:]=0.0
                return self.ratediv
        else:
            cols, delcols = self.spv.keys(), set()
            ratediv =np.zeros(self.nV, dtype=float)
            self.maxratediv = 0

        for col in cols:
            if col in delcols:
                assert(col not in self.spv)
                ratediv[col] = 0
                continue
            rs = self.spv[col]
            shis=np.bincount(rs, minlength=3)
            ahis = self.ahists[col]
            ohis=ahis-shis
            shis, ohis = shis+1, ohis+1 #a kind of multinomial prior
            if neutral is False:
                'remove netrual 2.5, 3, 3.5'
                shis[1], ohis[1] = 0, 0
            #cal KL-divergence
            from scipy import stats
            kl = stats.entropy(shis, ohis)
            lenrs = len(rs)
            lenars = len(self.apv[col])
            ssum, osum = float(lenrs)+1, float(lenars-lenrs)+1
            bal = (min(ssum/osum, osum/ssum))
            ratediv[col]=kl*bal
            self.maxratediv = kl if self.maxratediv < kl else self.maxratediv

        self.ratediv = ratediv
        return self.ratediv

    #@profile
    def suspburstinvolv(self, multiburstbd=0.5, weighted=True, delta=False):
        '''calc how many points allocated in awake and burst period, over total
           number of U who involv in the burst
        '''
        if delta and hasattr(self, 'incurstcnt') and hasattr(self, 'inburstratio'):
            cols, delcols = self.deltacols, self.delcols
            inburstcnt, inburstratio = self.inburstcnt, self.inburstratio
        else:
            inburstcnt, inburstratio = \
                    np.zeros(self.nV, dtype=int), np.zeros(self.nV, dtype=float)
            cols, delcols = self.spv.keys(), set()

        for col in cols:
            if col in delcols:
                assert(col not in self.spv)
                inburstcnt[col], inburstratio[col] = 0, 0.0
                continue
            st = self.spv[col]
            abpts, bvs, slops, ainburst = self.awakeburstpt[col], \
                    self.burstvals[col], self.burstslops[col], self.ainbursts[col]
            'get the satisfied multiburst points'
            burstids = bvs/float(bvs[0]) >=  multiburstbd
            abpts, slops, bvs, ainburst = abpts[burstids], slops[burstids],\
                    bvs[burstids], ainburst[burstids]
            scnt, wscnt, wallcnt =0, 0, 0
            for i in range(len(abpts)):
                (left, right), sp, bv, acnt = abpts[i],slops[i], bvs[i], ainburst[i]
                '#susp users in burst'
                cnt1 = ((st >= left) & (st <= right)).sum()
                scnt += cnt1
                '#all users in burst'
                assert(acnt>=cnt1)
                if weighted is not False:
                    wscnt +=  cnt1 * sp * bv
                    wallcnt += acnt * sp * bv
                else:
                    wscnt += cnt1
                    wallcnt += acnt
            inburstcnt[col]=scnt
            inburstratio[col] = wscnt/float(wallcnt)

        self.inburstcnt = inburstcnt
        self.inburstratio =inburstratio
        return self.inburstcnt, self.inburstratio

#@profile
def awakburstpoints_recur(ts, bins='auto'):
    'recursive version'
    hts = np.histogram(ts, bins=bins)
    ys = np.append([0], hts[0]) #add zero, so 0 is allocated to lowest left bound
    ys = ys.astype(np.float64)
    xs = hts[1]
    abptidxs = []
    startidx = 0
    'recursively get the idx for awake and burst pts'
    recurFindAwakePt(xs, ys, start=startidx, abptidxs=abptidxs)
    if len(abptidxs)==0:
        return [], [0], [0], None
    'extend left bound by -1, since we added zero to histogram'
    abptextidxs, bvsrt, slops = sort_extendLeftbd(abptidxs, xs, ys)
    'convert abptext idx to bd value in xs'
    abpts = np.array([(xs[l], xs[r]) for l, r in abptextidxs])
    return abpts, bvsrt, slops, [abptidxs, abptextidxs]

#@profile
def sort_extendLeftbd(abptidxs, xs, ys):
    'sort bds by burst val, and extend the left bound of sorted awakeburst pts'
    bv=[ ys[r]-ys[l] for l, r in abptidxs] #use abdiff as bv
    abptys = sorted(zip(abptidxs, bv), key=lambda x:x[1], reverse=True)
    abptsrt, bvsrt = zip(*abptys)
    abptsrt = np.array(abptsrt)
    bvsrt = np.array(bvsrt)
    'calculate slop of bursting before extending'
    slops, diffs = [], []
    for l, r in abptsrt:
        slop = (ys[r]-ys[l])/float(xs[r]-xs[l])
        slops.append(slop)
    slops = np.array(slops)
    #diffs = np.array(slops)
    'extend left, if overlep keep that of higher burst val'
    for i in range(len(abptsrt)):
        nl, nr = max(abptsrt[i][0]-1,0), abptsrt[i][1]
        for j in range(i):
            pl, pr = abptsrt[j][0], abptsrt[j][1]
            if nr >= pr and nl < pr:
                nl = pr
            if nl <= pl and nr > pl:
                print('[Warning] extended a impossible bound')
                nr = pl #impossible case, recurFindAwakePt guarantees that
        abptsrt[i][0], abptsrt[i][1]=nl,nr #extend
    return abptsrt, bvsrt, slops

#@profile
def recurFindAwakePt(xs, ys, start=0, abptidxs=[]):
    if len(ys)<=1 or len(xs)<=1:
        return
    maxidx = np.argmax(ys)
    x0,y0,xm,ym = xs[0], ys[0], xs[maxidx], ys[maxidx]
    sqco = math.sqrt((ym-y0)**2 + (xm-x0)**2) #sqrt of coefficient
    xvec, yvec = xs[:maxidx], ys[:maxidx]
    dts = ((ym-y0)*xvec - (xm-x0)*yvec + (xm*y0 - ym*x0))/sqco
    xaidx = np.argmax(dts)
    abptidxs.append((xaidx+start, maxidx+start))
    'left'
    recurFindAwakePt(xs[:xaidx], ys[:xaidx], start=start, abptidxs=abptidxs)
    'right'
    diffyincrese = np.argwhere(np.diff(ys[maxidx:]) >0)
    if len(diffyincrese) > 0:
        turningptidx = diffyincrese[0,0]+maxidx
        recurFindAwakePt(xs[turningptidx:], ys[turningptidx:],
                         start = turningptidx + start,
                         abptidxs=abptidxs)
    return

def burstmaxdying_recur(ts, endt, twait=12*3600, bins='auto'):
    'endt is used to judge if the dying is caused by observation window'
    #todo consider weights
    hts = np.histogram(ts, bins=bins)
    xs = hts[1]
    ys = hts[0].astype(np.float64)
    if  len(ys) < 2:
        return 0, xs[0], 0
    maxts = max(ts)
    if maxts < endt - twait:
        ys = np.concatenate((ys, [0.0]))
    else:
        hadd = (ys[-1]+ys[-2])/2.0
        ys = np.concatenate((ys, [hadd]))

    maxdying=[0.0, 0.0, 0.0]
    recurFindMaxFallDying(xs, ys, maxdying)
    return maxdying

def recurFindMaxFallDying(xs, ys, maxdying):
    lenys = len(ys)
    if lenys < 2:
        return
    burstidx = lenys - np.argmax(ys[::-1]) -1 #the last max occurrence
    if ys[burstidx]-min(ys) < maxdying[0]:
        return
    if burstidx == lenys-1: #bursting at the end
        dyingidx = lenys-1
        slop = (ys[burstidx] - 0)/float(xs[-1]-xs[-2])
        fall = ys[burstidx]
    else:
        xm, ym, xe, ye = xs[burstidx], ys[burstidx], xs[-1], ys[-1]
        sqco = math.sqrt((ym-ye)**2 + (xm-xe)**2) #sqrt of coefficient
        xvec, yvec = xs[burstidx+1:], ys[burstidx+1:]
        dts = -((ym-ye)*xvec - (xm-xe)*yvec + (xm*ye - ym*xe))/sqco
        dts = np.absolute(dts)
        dyingidx = len(dts)-np.argmax(dts[::-1])-1 + burstidx+1
        slop = (ym - ys[dyingidx])/float(xs[dyingidx] - xm)#dyingidx alwasy >0
        if dyingidx == lenys -1:
            fall = ym # assume continue to fall to 0, keeping current slop
        else:
            fall = ym-ys[dyingidx]
    if fall > maxdying[0]:
        maxdying[0:3] = [fall, xs[dyingidx], slop]

    if dyingidx < lenys-1:
        'move to the right'
        subburstidx = np.argmax(ys[dyingidx:]) + dyingidx
        recurFindMaxFallDying(xs[subburstidx:], ys[subburstidx:], maxdying)
    if burstidx > 1:
        'move to the left'
        subdyingidx = np.argmin(ys[:burstidx])
        recurFindMaxFallDying(xs[:subdyingidx], ys[:subdyingidx], maxdying)
    return


# assistant function. check tunit when is called
def pim2tensorformat(tsfile, ratefile, tensorfile, tunit='s', tbins='h'):
    'convert the pim files: tsfile, ratefile into tensor file, i.e. tuples'
    rbins = lambda x: 0 if x<2.5 else 1 if x<=3.5 else 2 #lambda x: x
    propdict = {}
    with myreadfile(tsfile, 'rb') as fts, myreadfile(ratefile, 'rb') as frt,\
            open(tensorfile, 'wb') as fte:
        for line in fts:
            k,v = line.strip().split(':')
            propdict[k]=[v]
        for line in frt:
            k,v=line.strip().split(':')
            propdict[k].append(v)
        for k, vs in propdict.items():
            u, b = k.strip().split('-')
            tss = vs[0].strip().split(' ')
            tss = list(map(int, tss))
            if tunit == 's':
                'time unit is second'
                if tbins == 'h':
                    'time bin size is hour'
                    tss = np.array(tss, dtype=int)//3600
                elif tbins == 'd':
                    'time bin size is day'
                    tss = np.array(tss, dtype=int)//(3600*24)
            'no matter what the tunit is'
            if type(tbins) is int:
                tss = np.array(tss, dtype=int)//tbins
            tss = list(map(str, tss))
            'process ts'
            rts = vs[1].strip().split(' ')
            rts = list(map(float, rts))
            digrs = []
            for r1 in rts:
                r = rbins(r1)
                digrs.append(r)
            digrs = list(map(int, digrs))
            digrs = list(map(str, digrs))
            for i in range(len(tss)):
                fte.write(','.join((u, b, tss[i], digrs[i], '1')))
                fte.write('\n')
        fts.close()
        frt.close()
        fte.close()
    return

# assistant function. check tunit when is called
def tspim2tensorformat(tsfile, tensorfile, tunit='s', tbins='h',
                       idstartzero=True):
    offset = 0 if idstartzero else -1
    propdict = {}
    with myreadfile(tsfile, 'rb') as fts, myreadfile(tensorfile, 'wb') as fte:
        for line in fts:
            k,v = line.strip().split(':')
            propdict[k]=[v]
        for k, vs in propdict.items():
            u, b = k.strip().split('-')
            if idstartzero is False:
                u = str(int(u)+offset)
                b = str(int(b)+offset)
            tss = vs[0].strip().split(' ')
            tss = list(map(int, tss))
            if tunit == 's':
                'time unit is second'
                if tbins == 'h':
                    'time bin size is hour'
                    tss = np.array(tss, dtype=int)//3600
                elif tbins == 'd':
                    'time bin size is day'
                    tss = np.array(tss, dtype=int)//(3600*24)
            if type(tbins) is int:
                tss = np.array(tss, dtype=int)//tbins
            tss = list(map(str, tss))
            for i in range(len(tss)):
                fte.write(','.join((u, b, tss[i], '1')))
                fte.write('\n')
        fts.close()
        fte.close()
    return






import numpy as np
from sklearn import metrics


def evaluate(labels, scores, set_th=None):

    auc_roc = metrics.roc_auc_score(labels, scores)
    auc_prc = metrics.average_precision_score(labels, scores)
    precisions, recalls, ths = metrics.precision_recall_curve(labels, scores)
    f1 = 2/(1/precisions+1/recalls)
    best_f1 = np.max(f1)
    best_th = ths[np.argmax(f1)]

    # threshold f1
    if set_th is not None:
        tmp_scores = scores.copy()
        tmp_scores[tmp_scores >= set_th] = 1
        tmp_scores[tmp_scores < set_th] = 0
        print(metrics.classification_report(labels, tmp_scores))
        print(metrics.confusion_matrix(labels, tmp_scores))

    return auc_prc, auc_roc, best_th, best_f1



import numpy as np
from scipy import sparse
from scipy.linalg import svd
import math
from spartan.tensor import DTensor


def generateGH_by_multiply(A, Omg):
    G = A.dot(Omg)
    H = A.T.dot(G)
    return G, H


def generateGH_by_list(G, H, glist, hlist, k):
    if k == 0:
        for g in glist:
            G = np.vstack((G, g))
        for h in hlist:
            H = H + h
    else:
        g = glist.pop(0)
        gm = g.shape[0]
        G = G[gm:, :]
        h = hlist.pop(0)
        H = H - h
        G = np.vstack((G, glist[-1]))
        H = H + hlist[-1]
    return G, H, glist, hlist


def generateQB(G, H, Omg, l, b):
    m = G.shape[0]
    n = H.shape[0]
    Q = np.zeros((m, 0))
    B = np.zeros((0, n))
    t = int(math.floor(l/b))
    for i in range(0, t):
        if B.shape[0] == 0:
            Yi = G[:, i*b: (i+1)*b]
            Qi, Ri = np.linalg.qr(Yi)
            Qi, Rit = np.linalg.qr(Qi)
            Ri = Rit * Ri
            invRi = DTensor.from_numpy(x=np.linalg.inv(Ri.T))
            Bi = invRi.dot(H[:, i * b: (i + 1) * b].T)
        else:
            temp = B.dot(Omg[:, i*b: (i+1)*b])
            Yi = G[:, i*b: (i+1)*b] - Q.dot(temp)
            Qi, Ri = np.linalg.qr(Yi)
            Qi, Rit = np.linalg.qr(Qi - Q.dot(Q.T.dot(Qi)))
            Ri = Rit * Ri
            invRi = DTensor.from_numpy(x=np.linalg.inv(Ri.T))
            dt_B, dt_Q = DTensor.from_numpy(B), DTensor.from_numpy(Q)
            Bi = invRi.dot(H[:, i*b: (i+1)*b].T - Yi.T.dot(dt_Q).dot(dt_B) - temp.T.dot(dt_B))
        Q = np.hstack((Q, Qi))
        B = np.vstack((B, Bi))
    return Q, B


def computeSVD(Q, B):
    u1, s, v = svd(B, full_matrices=False)
    u = Q.dot(u1)
    return u, s, v



import sys
import numpy as np
import random
import numpy.random as nr
import scipy.linalg as sla
# from .mytools.ioutil import *
from .mytools.ioutil import loadedge2sm, loadDictListData, loadDictListData, saveDictListData, savesm2edgelist, saveSimpleListData
from scipy.sparse import coo_matrix

def genEvenDenseBlock(A, B, p):
    m = []
    for i in range(A):
        a = np.random.binomial(1, p, B)
        m.append(a)
    return np.array(m)

def genHyperbolaDenseBlock(A, B, alpha, tau):
    'this is from hyperbolic paper: i^\alpha * j^\alpha > \tau'
    m = np.empty([A, B], dtype=int)
    for i in range(A):
        for j in range(B):
            if (i+1)**alpha * (j+1)**alpha > tau:
                m[i,j] = 1
            else:
                m[i,j] = 0
    return m

def genDiHyperRectBlocks(A1, B1, A2, B2, alpha=-0.5, tau=None, p=1):
    if tau is None:
        tau = A1**alpha * B1**alpha
    m1 = genEvenDenseBlock(A1, B1, p=p)
    m2 = genHyperbolaDenseBlock(A2, B2, alpha, tau)
    M = sla.block_diag(m1, m2)
    return M

def addnosie(M, A, B, p, black=True, A0=0, B0=0):
    v = 1 if black else 0
    for i in range(A-A0):
        a = np.random.binomial(1, p, B-B0)
        for j in a.nonzero()[0]:
            M[A0+i,B0+j]=v
    return M


def injectCliqueCamo(M, m0, n0, p, testIdx):
    (m,n) = M.shape
    M2 = M.copy().tolil()

    colSum = np.squeeze(M2.sum(axis = 0).A)
    colSumPart = colSum[n0:n]
    colSumPartPro = np.int_(colSumPart)
    colIdx = np.arange(n0, n, 1)
    population = np.repeat(colIdx, colSumPartPro, axis = 0)

    for i in range(m0):
        # inject clique
        for j in range(n0):
            if random.random() < p:
                M2[i,j] = 1
        # inject camo
        if testIdx == 1:
            thres = p * n0 / (n - n0)
            for j in range(n0, n):
                if random.random() < thres:
                    M2[i,j] = 1
        if testIdx == 2:
            thres = 2 * p * n0 / (n - n0)
            for j in range(n0, n):
                if random.random() < thres:
                    M2[i,j] = 1
        # biased camo
        if testIdx == 3:
            colRplmt = random.sample(population, int(n0 * p))
            M2[i,colRplmt] = 1

    return M2.tocsc()


def generateProps(rates, times, k, s, t0, tsdiffcands, tsp):

    if len(rates) > 0:
        rs = np.random.choice([4, 4.5], size=s)
        if k in rates:
            for r in rs:
                rates[k].append(r)
        else:
            rates[k] = list(rs)
    if len(times) > 0:
        ts = np.random.choice(tsdiffcands, size=s, p=tsp) + t0
        if k in times:
            for t in ts:
                times[k].append(t)
        else:
            times[k] = list(ts)
    return

def injectFraud2PropGraph(freqfile, ratefile, tsfile, acnt, bcnt, goal, popbd,
                          testIdx = 3, idstartzero=True, re=True, suffix=None,
                         weighted=True, output=True):
    if not idstartzero:
        print('we do not handle id start 1 yet for ts and rate')
        ratefile, tsfile = None, None

    # M = loadedge2sm(freqfile, coo_matrix, weighted=weighted, idstartzero=idstartzero)
    # remove unexpected parameter 'weighted'
    M = loadedge2sm(freqfile, coo_matrix, idstartzero=idstartzero)
    'smax: the max # of multiedge'
    smax = M.data.max() #max freqency
    if acnt == 0 and re:
        return M, ([], [])
    M2 = M.tolil()
    (m, n) = M2.shape
    rates, times, tsdiffs, t0 = [], [], [], 0
    t0, tsdiffcands,tsp = 0, [], []
    if ratefile is not None:
        rates = loadDictListData(ratefile, ktype=str, vtype=float)
    if tsfile is not None:
        times = loadDictListData(tsfile, ktype=str, vtype=int)
        tsmin, tsmax = sys.maxsize, 0
        tsdiffs = np.array([])
        prodts={i:[] for i in range(n)}
        for k,v in times.items():
            k = k.split('-')
            pid = int(k[1])
            prodts[pid] += v
        for pv in prodts.values():
            pv = sorted(pv)
            minv, maxv = pv[0], pv[-1]
            if tsmin > minv:
                tsmin = minv
            if tsmax < maxv:
                tsmax = maxv
            if len(pv)<=2:
                continue
            vdiff = np.diff(pv)
            'concatenate with [] will change value to float'
            tsdiffs = np.concatenate((tsdiffs, vdiff[vdiff>0]))
        tsdiffs.sort()
        tsdiffs = tsdiffs.astype(int)
        tsdiffcands = np.unique(tsdiffs)[:20] #another choice is bincount
        tsp = np.arange(20,dtype=float)+1
        tsp = 1.0/tsp
        tsp = tsp/tsp.sum()
        t0 = np.random.randint(tsmin, tsmax,dtype=int)

    colSum = M2.sum(0).getA1()
    colids = np.arange(n, dtype=int)
    targetcands = np.argwhere(colSum < popbd).flatten()
    targets = random.sample(targetcands, bcnt)
    camocands = np.setdiff1d(colids, targets, assume_unique=True)
    camoprobs = colSum[camocands]/float(colSum[camocands].sum())
    #population = np.repeat(camocands, colSum[camocands].astype(int), axis=0)
    fraudcands = np.arange(m,dtype=int) #users can be hacked
    fraudsters = random.sample(fraudcands, acnt)
    'rating times for one user to one product, multiedge'
    scands = np.arange(1,smax+1,dtype=int)
    sprobs = []
    numedges = (M>0).sum()
    for s in scands:
        nums = (M==s).sum()
        sprobs.append(float(nums)/numedges)

    # inject near clique
    for j in targets:
        exeusers = random.sample(fraudsters, goal)
        for i in exeusers:
            s = np.random.choice(scands, size=1, p=sprobs)[0] if weighted else 1
            if (not weighted) and M2[i,j] > 0:
                continue
            M2[i,j] += s
            k = '{}-{}'.format(i,j)
            generateProps(rates, times, k, s, t0, tsdiffcands,tsp)

    # inject camo
    p = goal/float(acnt)
    for i in fraudsters:
        if testIdx == 1:
            thres = p * bcnt / (n - bcnt)
            for j in camocands:
                s = np.random.choice(scands, size=1, p=sprobs) if weighted else 1
                if (not weighted) and M2[i,j] > 0:
                    continue
                if random.random() < thres:
                    M2[i,j] += s
                    k = '{}-{}'.format(i,j)
                    generateProps(rates, times, k, s, t0, tsdiffcands, tsp)
        if testIdx == 2:
            thres = 2 * p * bcnt / (n - bcnt)
            for j in camocands:
                s = np.random.choice(scands, size=1, p=sprobs) if weighted else 1
                if (not weighted) and M2[i,j] > 0:
                    continue
                if random.random() < thres:
                    M2[i,j] += s
                    k = '{}-{}'.format(i,j)
                    generateProps(rates, times, k, s, t0, tsdiffcands, tsp)
        # biased camo
        if testIdx == 3:
            colRplmt = np.random.choice(camocands, size=int(bcnt*p),
                                        p=camoprobs)
            #M2[i,colRplmt] = 1
            s = np.random.choice(scands, size=1, p=sprobs) if weighted else 1
            for j in colRplmt:
                if (not weighted) and M2[i,j] > 0:
                    continue
                M2[i,j] += s
                k = '{}-{}'.format(i,j)
                generateProps(rates, times, k, s, t0, tsdiffcands, tsp)

    if suffix is not None:
        suffix = str(suffix)
    else:
        suffix =''
    if ratefile is not None and output is True:
        saveDictListData(rates, ratefile+'.inject'+suffix)
    if tsfile is not None and output is True:
        saveDictListData(times, tsfile+'.inject'+suffix)
    M2 = M2.tocoo()
    if not weighted:
        M2.data[0:] =1
    if output is True:
        savesm2edgelist(M2.astype(int), freqfile+'.inject'+suffix, idstartzero=idstartzero)
        saveSimpleListData(fraudsters, freqfile+'.trueA'+suffix)
        saveSimpleListData(targets, freqfile+'.trueB'+suffix)
    if re:
        return M2, (fraudsters, targets)
    else:
        return




import numpy as np
import math


def filterEigenvec(A, u, v, k=0):
    m = A.shape[0]
    n = A.shape[1]
    num1 = 1 / math.sqrt(m)
    num2 = 1 / math.sqrt(n)
    list_u = u.tolist()
    list_v = v.tolist()
    list_i = [j for j, x in enumerate(list_u[k]) if abs(x) >= num1]
    list_j = [j for j, x in enumerate(list_v[k]) if abs(x) >= num2]
    B1 = A[list_i, :]
    B = B1[:, list_j]
    return B, list_i, list_j


def calDensity(mat):
    num = mat.sum()
    density = round(float(num) / (mat.shape[0] + mat.shape[1]), 2)
    return density


def findSuspWins(densities):
    mean = np.mean(densities)
    std = np.std(densities, ddof=1)
    thres = 3 * std + mean
    print('mean:{}, std val:{}, thres:{}'.format(mean, std, thres))
    burst_wins = []
    for i in range(densities.__len__()):
        if densities[i] > thres:
            burst_wins.append(i)
    return burst_wins


def getKeys(vs, dict):
    items, times = [], []
    for k in dict:
        v = dict[k]
        if v in vs:
            items.append(k[0])
            times.append(k[1])
    return items, times
        




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   drawutil.py
@Desc    :   Draw functions.
'''

# here put the import lib
import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from spartan.tensor import Graph, Timeseries

# TODO do not import matplotlib function in model file


def plot(model: "Model", *args, **kwargs):
    from spartan import model as _model
    function_dict = {
        _model.BeatLex: __plot_beatlex,
        _model.BeatGAN: __plot_beatgan
    }
    if function_dict.__contains__(model):
        function_dict[model](*args, **kwargs)
    else:
        raise Exception(f"Draw functions of model {model} not existed")


def plot_graph(graph: Graph, layout=None, bipartite=False, labels=None,
               *args, **kwargs):
    import networkx as nx
    from networkx.algorithms.bipartite.matrix import from_biadjacency_matrix
    nx_layout = {
        None: nx.random_layout,
        'circular': nx.circular_layout,
        'kamada_kawai': nx.kamada_kawai_layout,
        'random': nx.random_layout,
        'shell': nx.shell_layout,
        'spectral': nx.spectral_layout,
        'spring': nx.spring_layout,
        'bipartite': nx.bipartite_layout
    }

    nrow, ncol = graph.sm.shape
    if bipartite:
        g = from_biadjacency_matrix(graph.sm)
        # bipartite graph can use other layout, but default is bipartite
        layout = 'bipartite' if layout is None else layout
    else:
        g = nx.from_scipy_sparse_matrix(graph.sm)

    if layout is 'bipartite':
        pos = nx.bipartite_layout(g, nodes=range(nrow))
    else:
        pos = nx_layout[layout](g)

    fig = plt.figure()
    if bipartite:
        nx.draw_networkx(g, pos=pos,
                         node_color=('r'*nrow+'b'*ncol), alpha=0.8)
    else:
        nx.draw_networkx(g, pos=pos, node_color='r', alpha=0.8)

    if labels is not None:
        if isinstance(labels, dict):
            nx.draw_networkx_labels(g, pos=pos, labels=labels)
        else:
            ldict = dict(zip(range(len(labels)), labels))
            nx.draw_networkx_labels(g, pos=pos, labels=ldict)

    return fig


def plot_timeseries(*args, **kwargs):
    import matplotlib.pyplot as plt
    __plot_timeseries(plt, *args, **kwargs)
    plt.show()


def __plot_timeseries(plt, series: Timeseries, chosen_labels: list = None):
    if chosen_labels is None:
        sub_dimension = series.dimension
        actual_dimension = 1
        fig, ax = plt.subplots(sub_dimension, 1, sharex=True)
        if type(ax) is not np.ndarray:
            ax = [ax]
        for index, label in enumerate(series.labels):
            ax[index].set_title(label)
            ax[index].plot(series.time_tensor._data, series.val_tensor._data[index], label=label)
            ax[index].legend(loc="best")
            actual_dimension += 1
    else:
        sub_dimension = len(chosen_labels)
        actual_dimension = 1
        fig, ax = plt.subplots(sub_dimension, 1, sharex=True)
        if type(ax) is not np.ndarray:
            ax = [ax]
        for chosen_label in chosen_labels:
            for label in chosen_label:
                index = series.labels.index(label)
                ax[actual_dimension-1].plot(series.time_tensor._data, series.val_tensor._data[index], label=label)
            ax[actual_dimension-1].legend(loc="best")
            ax[actual_dimension-1].set_title(', '.join(chosen_label))
            actual_dimension += 1
    plt.xlabel('time/s')


def plot_resampled_series(series: Timeseries, origin_length: int, resampled_length: int, origin_freq: int, resampled_freq: int, origin_list, resampled_list, start):
    plt.figure()
    sub_dimension = len(resampled_list)
    actual_dimension = 1
    for label in series.labels:
        x_origin = np.arange(0, origin_length/origin_freq, 1/origin_freq)
        x_resampled = np.arange(0, resampled_length/resampled_freq, 1/resampled_freq)
        x_origin += start
        x_resampled += start
        plt.subplot(sub_dimension, 1, actual_dimension)
        index = series.labels.index(label)
        plt.title(label)
        plt.plot(x_origin, origin_list[index], 'r-', label='origin')
        plt.plot(x_resampled, resampled_list[index], 'g.', label='resample')
        plt.legend(loc="best")
        actual_dimension += 1
    plt.xlabel('time/s')
    plt.show()


def drawEigenPulse(densities: list = [], figpath: str = None):
    xs = range(len(densities))
    plt.plot(xs, densities, label='density')
    plt.xlabel('window idx')
    plt.ylabel('density')

    thres = np.mean(densities) + 3 * np.std(densities)
    plt.hlines(thres, min(xs), max(xs), linestyles='dashed',
               colors='yellow', label='threshold')
    plt.legend()
    if figpath is None:
        plt.show()
    else:
        plt.savefig(figpath)


def __plot_beatlex(time_series, result):
    import matplotlib.pyplot as plt
    models = result['models']
    plt = __plot_beatlex_vocabulary(plt, models)
    plt = __plot_beatlex_timeseries(plt, time_series, result)
    plt.show()
    return plt


def __plot_beatlex_vocabulary(plt, models):
    color = ['r', 'g', 'b', 'c'] * (int(len(models) / 4)+1)
    for i, model in enumerate(models):
        plt.figure()
        plt.plot(model.T, color[i % len(models)])
        plt.title(f"Vocabulary {i}")
    return plt


def __plot_beatlex_timeseries(plt, series: Timeseries, result: dict):
    starts = result['starts']
    ends = result['ends']
    model_num = len(result['models'])
    idx = result['idx']
    color = ['r', 'g', 'b', 'c'] * (int(model_num / 4)+1)
    sub_dimension = series.dimension
    actual_dimension = 1
    _, ax = plt.subplots(sub_dimension, 1, sharex=True)
    if sub_dimension == 1:
        ax = [ax]
    for index, label in enumerate(series.labels):
        ax[index].set_title(label)
        if starts is not None and ends is not None:
            for i, _start in enumerate(starts):
                _end = ends[i]
                ax[index].plot(series.time_tensor._data[_start: _end], series.val_tensor._data[index][_start:_end], color[idx[i]], label=label if i == 0 else None)
                ax[index].scatter(series.time_tensor._data[_start], series.val_tensor._data[index][_start], color='black')
        else:
            ax[index].plot(series.time_tensor._data, series.val_tensor._data[index], label=label)
        ax[index].legend(loc="best")
        actual_dimension += 1
    plt.xlabel('time/s')
    return plt


def __plot_beatgan(input, output, heat):
    import matplotlib.pyplot as plt
    sig_in = input.transpose()
    sig_out = output.transpose()
    max_heat = np.max(heat)
    min_heat = np.min(heat)
    x_points = np.arange(sig_in.shape[0])
    fig, ax = plt.subplots(2, 1, sharex=True, figsize=(6, 6), gridspec_kw={'height_ratios': [6, 1]})
    ax[0].plot(x_points, sig_in, 'k-', linewidth=2.5, label="input signal")
    ax[0].plot(x_points, sig_out, 'k--', linewidth=2.5, label="output signal")
    ax[0].legend(loc="upper right")
    heat = (sig_out-sig_in)**2
    heat_norm = (heat-min_heat)/(max_heat-min_heat)
    heat_norm = np.reshape(heat_norm, (1, -1))
    ax[1].imshow(heat_norm, cmap="jet", aspect="auto")
    ax[1].set_yticks([])
    fig.tight_layout()


def histogram_viz(histogram_matrix, x_ticks: list, y_ticks: list, output: str,
                  base: int = 10, x_label: str = None, y_label: str = None):
    '''Plot two-dimensional histogram
    Parameters:
    --------
    :param histogram_matrix: sparse matrix
        histogram data represented with sparse matrix format
    :param x_ticks: list
        ticks of x-axis
    :param y_ticks: list
        ticks of x-axis
    :param base: int
        the logarithmic base for bucketing the graph features.
        Default is 10.
    :param output: str
        output path of the figure
    :param x_label: str
        label of x-axis
        Default is None
    :param y_label: str
        label of y-axis
        Dedault is None
    '''
    import matplotlib.pylab as plt
    from matplotlib.colors import LogNorm

    n, m = histogram_matrix.shape
    fig = plt.figure(figsize=(4.7, 3.8), dpi=96)  # figsize=(8, 6.5),
    plt.pcolormesh(histogram_matrix.toarray(), cmap='jet', norm=LogNorm(), rasterized=True)
    cb = plt.colorbar()
    cb.set_label('counts')
    # for lb in cb.ax.yaxis.get_ticklabels():
    # lb.set_family('Times New roman')
    # lb.set_size(20)

    ax = fig.gca()
    xticks = ax.get_xticks()
    if xticks[-1] > m:
        xticks = xticks[:-1]
    xstep = xticks[1] - xticks[0]
    nw_xtick = []
    for xt in xticks:
        if (xt < m) and (xt % (2*xstep) == 0):
            pws = int(np.log(x_ticks[int(xt)]) / np.log(base))
            if pws != 0:
                nw_xtick.append('%dE%d' % (base, pws))
            else:
                nw_xtick.append('1')
        else:
            nw_xtick.append('')

    nw_ytick = []
    for yt in ax.get_yticks():
        if yt < n:
            yval = y_ticks[int(yt)]
            if yval < 1e4:
                nw_ytick.append(r'%d' % yval)
            else:
                pws = int(np.log10(yval))
                fv = yval * 1.0 / 10**pws
                nw_ytick.append('%.1fE%d' % (fv, pws))

    if nw_xtick[-1] == '':
        nw_xtick[-1] = '%.2f' % x_ticks[-1]
        # nw_xtick[-1] = '%.2f'%np.power(base, x_vec[-1])
    if nw_ytick[-1] == '':
        nw_ytick[-1] = '%d' % int(y_ticks[-1])
        # nw_ytick = '%d' % int(np.power(base, y_vec[-1]))

    ax.set_xticklabels(nw_xtick)  # , fontsize=27, family='Times New roman'
    ax.set_yticklabels(nw_ytick)  # , fontsize=27, family='Times New roman'

    if x_label is not None:
        plt.xlabel(x_label, linespacing=12)  # , fontsize=32, family='Times New roman'
    if y_label is not None:
        plt.ylabel(y_label, linespacing=12)  # , fontsize=32, family='Times New roman'

    # fig.set_size_inches(8, 7.3)
    fig.tight_layout()
    if output is not None:
        fig.savefig(output)
    return fig


def __plot_cluster(data, center_pts, data_labels, outliers=list(),
                   core_samples=None, grid=False, outfn=None):
    '''plot clusters of the data'''
    import collections as clct
    import matplotlib.cm as cm
    import matplotlib.pylab as plt

    fig = plt.figure(figsize=(3.8, 3.7), dpi=96)  # figsize=(6.5, 6)
    lab2cnt = clct.Counter(data_labels)
    cmap = cm.get_cmap('Spectral')
    colors = cmap(np.linspace(0, 1, len(lab2cnt)))
    if core_samples is not None:
        core_samples_mask = np.zeros_like(data_labels, dtype=bool)
        core_samples_mask[core_samples] = True
    else:
        core_samples_mask = np.ones_like(data_labels, dtype=bool)

    N_clusters = 0
    keys, values = np.array(list(lab2cnt.keys())), np.array(list(lab2cnt.values()))
    srt_lab = keys[np.argsort(values)][::-1]
    for k, col in zip(srt_lab, colors):
        if k == -1:
            # Black used for noise.
            # col = 'gray'
            continue

        N_clusters += 1
        class_member_mask = (data_labels == k)
        xy = data[class_member_mask & core_samples_mask]
        if len(xy) > 0:
            plt.plot(xy[:, 1], xy[:, 0], 's', color=col, markersize=6)  # markerfacecolor=col, markeredgecolor=col

        mn_xy = np.mean(xy, 0)
        plt.text(mn_xy[1], mn_xy[0], str(k),
                 {'color': 'k', 'fontsize': 18, 'ha': 'center', 'va': 'center',
                  'bbox': dict(boxstyle="circle", fc="w", ec="k", pad=0.2, alpha=0.3)})
        xy = data[class_member_mask & (np.invert(core_samples_mask))]
        if len(xy) > 0:
            plt.plot(xy[:, 1], xy[:, 0], 's', color=col, markersize=6)  # markerfacecolor=col, markeredgecolor=col
        if len(center_pts) > 0:
            plt.plot(center_pts[k, 0], center_pts[k, 1], 's', markerfacecolor=col, markeredgecolor='k', markersize=10)

    ubnd, rbnd = np.max(np.vstack((data, outliers)), 0) + 1
    if len(outliers) > 0:
        plt.plot(outliers[:, 1], outliers[:, 0], 'bo', markersize=1)
        uo, ro = np.max(outliers, 0)
        rbnd, ubnd = np.max([rbnd, ro]), np.max([ubnd, uo])

    ax = plt.gca()
    ax.tick_params(axis='both', which='both', bottom='off', left='off', labelbottom='off', labelleft='off', )
    # ax.set_axis_bgcolor('white')   # deprecated method in Matplotlib v2.0
    ax.set_facecolor('white')

    spine_linewidth = 3  # 6 #8#
    for spine in ax.spines.values():
        spine.set_linewidth(spine_linewidth)
    plt.grid(grid)
    plt.xlim((0, rbnd + 0.2))
    plt.ylim((0, ubnd + 0.2))
    plt.tight_layout()
    if outfn is not None:
        plt.savefig(outfn)
        plt.close()
    # plt.show()
    return fig


def clusters_viz(hcel2label: dict, output: str, outlier_label=-1):
    '''Visualize cluster result for histogram
    Parameters:
    -------
    :param hcel2label: dict
        histogram cell to its label
    :param output: str
        output path of the figure
    :param outlier_label: int
        label of the outliers for the 'hcel2label' data.
        Default is -1
    '''
    import warnings
    warnings.filterwarnings("ignore")

    def size_relabel(labels):
        clsdic = {}
        for l in labels:
            if l not in clsdic:
                if l != -1:
                    clsdic[l] = len(clsdic)
                else:
                    clsdic[l] = l
        rlbs = [clsdic[l] for l in labels]
        return np.array(rlbs)

    hcel_lab = np.column_stack((list(hcel2label.keys()), list(hcel2label.values())))
    outs_idx = hcel_lab[:, -1] == outlier_label
    outs = hcel_lab[outs_idx, :-1]
    others, others_lab = hcel_lab[~outs_idx, :-1], hcel_lab[~outs_idx, -1]
    others_lab = size_relabel(others_lab)
    cls_fig = __plot_cluster(others, [], others_lab, outliers=outs[::-1])
    if output is not None:
        cls_fig.savefig(output)
    return cls_fig


def drawHexbin(xs, ys, outfig=None, xscale='log', yscale='log',
               gridsize=200,
               colorscale=True,
               suptitle='Hexagon binning points',
               xlabel='', ylabel=''):
    '''
        xscale: [ 'linear' | 'log' ]
            Use a linear or log10 scale on the horizontal axis.
    yscale: [ 'linear' | 'log' ]
            Use a linear or log10 scale on the vertical axis.
        gridsize: [ 100 | integer ]
            The number of hexagons in the x-direction, default is 100. The
            corresponding number of hexagons in the y-direction is chosen such that
            the hexagons are approximately regular. Alternatively, gridsize can be
            a tuple with two elements specifying the number of hexagons in the
            x-direction and the y-direction.
    '''
    xs = np.array(xs) if type(xs) is list else xs
    ys = np.array(ys) if type(ys) is list else ys
    if xscale == 'log' and min(xs) <= 0:
        print('[Warning] logscale with nonpositive values in x coord')
        print('\tremove {} nonpositives'.format(len(np.argwhere(xs <= 0))))
        xg0 = xs > 0
        xs = xs[xg0]
        ys = ys[xg0]
    if yscale == 'log' and min(ys) <= 0:
        print('[Warning] logscale with nonpositive values in y coord')
        print('\tremove {} nonpositives'.format(len(np.argwhere(ys <= 0))))
        yg0 = ys > 0
        xs = xs[yg0]
        ys = ys[yg0]

    fig = plt.figure()
    if colorscale:
        plt.hexbin(xs, ys, bins='log', gridsize=gridsize, xscale=xscale,
                   yscale=yscale, mincnt=1, cmap=plt.cm.jet)
    else:
        plt.hexbin(xs, ys, gridsize=gridsize, xscale=xscale, yscale=yscale,
                   mincnt=1, cmap=plt.cm.jet)

    suptitle = suptitle + ' with a log color scale' if colorscale else suptitle
    plt.title(suptitle)

    cb = plt.colorbar()
    if colorscale:
        cb.set_label('log10(N)')
    else:
        cb.set_label('counts')
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if outfig is not None:
        fig.savefig(outfig)
    return fig


def drawRectbin(xs, ys, outfig=None, xscale='log', yscale='log',
                gridsize=200,
                colorscale=True,
                suptitle='Rectangle binning points',
                xlabel='', ylabel=''):
    xs = np.array(xs) if type(xs) is list else xs
    ys = np.array(ys) if type(ys) is list else ys
    if xscale == 'log' and min(xs) <= 0:
        print('[Warning] logscale with nonpositive values in x coord')
        print('\tremove {} nonpositives'.format(len(np.argwhere(xs <= 0))))
        xg0 = xs > 0
        xs = xs[xg0]
        ys = ys[xg0]
    if yscale == 'log' and min(ys) <= 0:
        print('[Warning] logscale with nonpositive values in y coord')
        print('\tremove {} nonpositives'.format(len(np.argwhere(ys <= 0))))
        yg0 = ys > 0
        xs = xs[yg0]
        ys = ys[yg0]

    fig = plt.figure()

    # color scale
    cnorm = matplotlib.colors.LogNorm() if colorscale else matplotlib.colors.Normalize()
    suptitle = suptitle + ' with a log color scale' if colorscale else suptitle

    # axis space
    if isinstance(gridsize, tuple):
        xgridsize = gridsize[0]
        ygridsize = gridsize[1]
    else:
        xgridsize = ygridsize = gridsize
    if xscale == 'log':
        xlogmax = np.ceil(np.log10(max(xs)))
        x_space = np.logspace(0, xlogmax, xgridsize)
    else:
        x_space = xgridsize
    if yscale == 'log':
        ylogmax = np.ceil(np.log10(max(ys)))
        y_space = np.logspace(0, ylogmax, ygridsize)
    else:
        y_space = ygridsize

    hist = plt.hist2d(xs, ys, bins=(x_space, y_space), cmin=1, norm=cnorm,
                      cmap=plt.cm.jet)
    plt.xscale(xscale)
    plt.yscale(yscale)

    plt.title(suptitle)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    cb = plt.colorbar()
    if colorscale:
        cb.set_label('log10(N)')
    else:
        cb.set_label('counts')

    if outfig is not None:
        fig.savefig(outfig)
    return fig, hist


def plot_tri_partitate(quad_tuples, a_ids, m_ids, c_ids, highlight_ids = None):
    '''
    plot for tri-partitate graph.

    Args:
        highlight_ids: dict. key - value is like (color, ids)

    Note: 
        id of a, m and c must be unique.
    '''
    if isinstance(a_ids, set) or isinstance(m_ids, set) or isinstance(c_ids, set):
        raise ValueError("ids of a, m and c must be a set.")


    BLUE = '#A0CBE2'
    INF = 999999999
    G = nx.Graph()
    dict_in = dict()
    dict_out = dict()
 

    # compute statistics value
    for u, v, w, c in quad_tuples:
        G.add_edge(u, v, weight = w, count = c)
        dict_out[u] = dict_out.get(u, 0) + w
        dict_in[v] = dict_in.get(v, 0) + w

    # node colors
    node_colors = [BLUE] * G.number_of_nodes()
    if highlight_ids is not None:
        import numpy as np
        node_ids = np.array(list(G.nodes()), dtype=int)
        ids_map = node_ids.argsort()
        # print(node_ids)
        # print(ids_map)
        for color, ids in highlight_ids.items():
            for idx in ids:
                node_colors[ids_map[idx-1]] = color

    # node positions
    components = [list(G.subgraph(c).nodes()) for i, c in enumerate(nx.connected_components(G))]
    node_list = [i for p in components for i in p]

    idx1 = [a for a in node_list if a in a_ids]
    idx2 = [m for m in node_list if m in m_ids]
    idx3 = [c for c in node_list if c in c_ids]

    pos = dict()
    max_size = max(len(idx1), len(idx2), len(idx3))
    pos.update((n, (0, i*3.0*max_size / len(idx1))) for i, n in enumerate(idx1))
    pos.update((n, (1, i*3.0*max_size / len(idx2))) for i, n in enumerate(idx2))
    pos.update((n, (2, i*3.0*max_size / len(idx3))) for i, n in enumerate(idx3))

    options = {'node_color':node_colors, 'node_size':500, 'width':0.5, 'pos':pos, 'with_labels':True, 'edge_color':BLUE}
    edge_labels = nx.get_edge_attributes(G, 'weight')
    count = nx.get_edge_attributes(G, 'count')
    for k, v in count.items():
        edge_labels[k] = "%s, %s" % (edge_labels[k], v)
    
    # node labels
    node_labels = dict((x, x) for x in G.nodes())
    for m in idx2:
        node_labels[m] = "%s\nin_%d_out_%d_q-f_%d" % (node_labels[m], dict_in[m], 
                                dict_out[m], abs(dict_in[m]-dict_out[m]))
    for a in idx1:
        node_labels[a] = "%s\nout_%d" % (node_labels[a], dict_out[a])
    for c in idx3:
        node_labels[c] = "%s\nin_%d" % (node_labels[c], dict_in[c])

    plt.figure(figsize=(10, 8))
    nx.draw(G, labels = node_labels, **options)
    nx.draw_networkx_edge_labels(G, pos, edge_labels)
    plt.show()



import sys
sys.path.append('../')
from gendenseblock import *
from mytools.ioutil import saveSimpleListData
import numpy as np
import scipy.sparse.linalg as slin
from scipy.sparse import csr_matrix
from gendenseblock import *
from os.path import expanduser
import math
home = expanduser("~")

def svddenseblock(m, rbd='avg'):
    m = m.asfptype()
    u, s, vt = slin.svds(m, k=1, which='LM')
    u1 = u[:,0]
    v1 = vt[0,:]
    s1 = s[0]
    if abs(max(u1)) < abs(min(u1)):
        u1 = -1*u1
    if abs(max(v1)) < abs(min(v1)):
        v1 = -1*v1
    sqrtS1 = math.sqrt(s1)
    if type(rbd) is float:
       rows = ((sqrtS1*u1)>=rbd).astype(int)
       cols = ((sqrtS1*v1)>=rbd).astype(int)
    elif rbd == 'avg':
       nrow, ncol = m.shape
       rows = (u1>=1.0/math.sqrt(nrow)).astype(int)
       cols = (v1>=1.0/math.sqrt(ncol)).astype(int)
    #rows = np.round(sqrtS1*u1).astype(int)
    #cols = np.round(sqrtS1*v1).astype(int)
    return rows, cols

def svddenseblockrank(m):
    m = m.asfptype()
    u, s, vt = slin.svds(m, k=1, which='LM')
    u1 = u[:,0]
    v1 = vt[0,:]
    s1 = s[0]
    if abs(max(u1)) < abs(min(u1)):
        u1 = -1*u1
    if abs(max(v1)) < abs(min(v1)):
        v1 = -1*v1
    sqrtS1 = math.sqrt(s1)
    rows = sqrtS1*u1
    cols = sqrtS1*v1
    return rows, cols

if __name__=="__main__":
    datapath=home+'/Data/'
    testdatapath='./testdata/'
    respath='./testout/'
    dataname='example.txt'
    data=testdatapath+dataname
    #coom = readedge2coom(data, weighted=False, idstartzero=True)
    print('loading data ... ...')

    '''
    M = genTriDenseBlock(1000, 1000, 1000, 500, 1000,1000, p1=0.8, alpha2=3,
                         alpha3=9.0)
    sm=csr_matrix(M)
    '''
    '''
    m = genDiDenseBlock(500,500, 1, 1500, 1500, alpha=-1)
    m=addnosie(m, 2000, 2000, 0.4, black=False)
    m=addnosie(m, 2000, 500, 0.005, black=True, A0=500, B0=0)
    m=addnosie(m, 500, 2000, 0.005, black=True, A0=0, B0=500)
    '''
    #m = genTriRectBlocks(3000,3000,0.6,0.6,0.6)
    #m = genDiHyperRectBlocks(50, 50, 2500, 2500, alpha=-0.5)
    '''when hiperbola: contains 50x50 core, rect: 88x88 can be
    detected by genDiHyperRectBlocks (optimal)
       rect: < 88x88 will not be detected, instead it detects the 58x58 in
       hyperbolia
    '''
    A1,B1,A2,B2= 500,500, 2500, 2500 #100,100, 2500, 2500 #88, 88, 2500, 2500,
    m = genDiHyperRectBlocks(A1, B1, A2, B2, alpha=-0.5, tau=0.002)
    m = addnosie(m, A1+A2, B1+B2, 0.005, black=True, A0=0, B0=0)
    m = addnosie(m, A1+A2, B1+B2, 0.4, black=False, A0=0, B0=0)
    sm=csr_matrix(m, dtype=np.float64)
    drows, dcols = svddenseblock(sm)
    print("dense block size: {}x{}".format(len(drows.nonzero()[0]),
                                           len(dcols.nonzero()[0])))




#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  Statistic hypothesis Test for truncate normal distribution
#  Author: wenchieh
#
#  Project: eaglemine
#      statistic_hypothesis_test_truncate.py
#      Version:  1.0
#      Date: December 1 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/1/2017>
#

__author__ = 'wenchieh'

# third-party lib
import numpy as np

# project
from .truncate_norm_gof import TruncatedNormalGoF


class StatisticHypothesisTestTruncate(object):
    def __init__(self, alpha_level=0.1, n_jobs=1):
        self.alpha = alpha_level
        self._n_jobs_ = n_jobs
        self._min_float = np.finfo('float64').resolution

    def _expand_(self, xs, count):
        if np.max(count) == 1:
            return np.array(xs)
        else:
            exp_xs = list()
            for k in range(len(xs)):
                exp_xs.extend([list(xs[k])] * count[k])
            return np.asarray(exp_xs)

    def _mixture_assign_prob_(self, xs, xs_probs, components, mix_weights):
        assigns = list()
        xs_wprob = np.array(xs_probs) * np.array(mix_weights)
        xs_comp_ = np.argmax(xs_wprob, axis=1)
        for k in range(components):
            assigns.append(np.where(xs_comp_ == k)[0])

        return assigns

    def _trunc_gaussian_check_(self, ndim, xs, means, stds, Hs, verbose=False):
        trunc_gaus = True
        gofer = TruncatedNormalGoF(self._n_jobs_)
        for k in range(ndim):
            # k_res = gofer.ad2_test(xs[:, k], means[k], stds[k], Hs[k])
            k_res = gofer.ad2up_test(xs[:, k], means[k], stds[k], Hs[k])
            # k_res = gofer.adup_test(xs[:, k], means[k], stds[k], Hs[k], 'greater')
            k_TS, k_pvalue = k_res['statistic'], k_res['p_value']
            if verbose:
                print("dim: {}, TS: {}, p-value:{}".format(k, k_TS, k_pvalue))
            if k_pvalue < self.alpha:
                trunc_gaus = False
                break
        return trunc_gaus

    def apply(self, xs, xs_weights, xs_probs,  paras, is_mix, Hs):
        xs = np.asarray(xs)
        xs = self._expand_(xs, np.array(np.ceil(xs_weights), int))
        m, n = xs.shape
        if is_mix:
            mus, covs, mix_wt = paras.get("mus"), paras.get("covs"), paras.get("weights")
            n_mixs = len(mus)
            xs_probs = self._expand_(xs_probs, np.array(np.ceil(xs_weights), int))
            assigns = self._mixture_assign_prob_(xs, xs_probs, n_mixs, mix_wt)
            mix_trunc_gaus = True
            for i in range(n_mixs):
                xs_imix = xs[assigns[i], :]
                mix_trunc_gaus &= self._trunc_gaussian_check_(n, xs_imix, mus[i],
                                                              np.sqrt(np.diag(covs[i])), Hs)
                if not mix_trunc_gaus:
                    break
            return mix_trunc_gaus

        else:
            means, stds = paras.get("mus")[0], np.sqrt(np.diag(paras.get("covs")[0]))
            return self._trunc_gaussian_check_(n, xs, means, stds, Hs)


import warnings

import numpy as np
from datasketch import MinHash
from datasketch.hashfunc import sha1_hash32

_mersenne_prime = (1 << 61) - 1
_max_hash = (1 << 32) - 1
_hash_range = (1 << 32)

class NeiMinHash(MinHash):
    def __init__(self, num_perm=128, seed=1,
            hashfunc=sha1_hash32,
            hashobj=None, # Deprecated.
            hashvalues=None, permutations=None):
        if hashvalues is not None:
            num_perm = len(hashvalues)
        if num_perm > _hash_range:
            # Because 1) we don't want the size to be too large, and
            # 2) we are using 4 bytes to store the size value
            raise ValueError("Cannot have more than %d number of\
                    permutation functions" % _hash_range)
        self.seed = seed
        # Check the hash function.
        if not callable(hashfunc):
            raise ValueError("The hashfunc must be a callable.")
        self.hashfunc = hashfunc
        # Check for use of hashobj and issue warning.
        if hashobj is not None:
            warnings.warn("hashobj is deprecated, use hashfunc instead.",
                    DeprecationWarning)
        # Initialize hash values
        if hashvalues is not None:
            self.hashvalues = self._parse_hashvalues(hashvalues)
        else:
            self.hashvalues = self._init_hashvalues(num_perm)
        # Initalize permutation function parameters
        if permutations is not None:
            self.permutations = permutations
        else:
            generator = np.random.RandomState(self.seed)
            # Create parameters for a random bijective permutation function
            # that maps a 32-bit hash value to another 32-bit hash value.
            # http://en.wikipedia.org/wiki/Universal_hashing
            a = generator.randint(1, _mersenne_prime, num_perm, dtype=np.uint64)
            b = generator.randint(0, _mersenne_prime, num_perm, dtype=np.uint64)
            self.a, self.b = a, b.reshape(len(b), 1)
            self.permutations = np.vstack([a, b])
        if len(self) != len(self.permutations[0]):
            raise ValueError("Numbers of hash values and permutations mismatch")

    def updates(self, values):
        values = np.array(values, dtype=np.uint64)

        phv = (np.outer(self.a, values) + self.b) % _mersenne_prime
        phv = np.bitwise_and(phv, np.uint64(_max_hash)).min(axis=1)
        self.hashvalues = np.minimum(self.hashvalues, phv, dtype=np.uint64)


import importlib
import json
import os
import sys
import warnings

from .cpu_backend import STensor, DTensor

__all__ = [
    'get_backend', 'get_preferred_backend', 'load_backend',
    'set_default_backend', 'STensor', 'DTensor'
]

# Inconsistent apis: sort(), round()
_APIS = [
    'add', 'all', 'angle', 'any', 'arange', 'argmax', 'argmin', 'argsort',
    'bincount', 'bitwise_and', 'bitwise_not', 'bitwise_or', 'bitwise_xor',
    'can_cast', 'ceil', 'complex128', 'complex64', 'conj', 'concatenate',
    'cos', 'cosh', 'cumprod', 'cumsum', 'diag', 'diagflat', 'diagonal',
    'dot', 'empty', 'empty_like', 'equal', 'exp', 'expm1', 'eye', 'flip',
    'float16', 'float32', 'float64', 'floor', 'floor_divide', 'fmod', 'full',
    'full_like', 'imag', 'int16', 'int32', 'int64', 'int8', 'isfinite',
    'isinf', 'isnan', 'linspace', 'log', 'log10', 'log1p', 'log2',
    'logical_and', 'logical_not', 'logical_or', 'logical_xor', 'logspace',
    'matmul', 'mean', 'median', 'meshgrid', 'nonzero', 'ones', 'ones_like',
    'prod', 'real', 'reciprocal', 'remainder', 'reshape', 'roll', 'rot90',
    'short', 'sign', 'sin', 'sinh', 'split', 'sqrt', 'square', 'squeeze',
    'stack', 'std', 'sum', 'take', 'tan', 'tanh', 'tensordot', 'trace',
    'transpose', 'tril', 'tril_indices', 'triu', 'triu_indices', 'true_divide',
    'trunc', 'uint8', 'unique', 'var', 'where', 'zeros', 'zeros_like',
    'STensor', 'DTensor'
]

_backend_dict = {
    'cpu': 'cpu_backend',
    'gpu': 'gpu_backend'
}


_BACKEND = 'cpu'


def load_backend(backend_name):
    """Load specific backend for spartan.

    Parameters
    ----------
    backend_name : str, {'cpu', 'gpu'}
        Backend name
    """
    global _BACKEND
    _BACKEND = backend_name.lower()
    if _BACKEND not in _backend_dict:
        msg = f"Unsupported backend: {_BACKEND}. Use CPU backend instead."
        warnings.warn(msg)
        _BACKEND = 'cpu'
    # Check GPU
    if _BACKEND == 'gpu':
        mod = importlib.import_module('torch')
        if not mod.cuda.is_available():
            msg = "GPU not available! Use CPU backend instead."
            warnings.warn(msg)
            _BACKEND = 'cpu'

    print(f"Using backend {_BACKEND}")
    thismod = sys.modules['spartan']
    mod_name = '.' + _backend_dict[_BACKEND]
    mod = importlib.import_module(mod_name, __name__)
    for api in _APIS:
        if api in mod.__dict__:
            setattr(thismod, api, mod.__dict__[api])
        else:
            msg = f"'{api}' not implemented in '{_BACKEND}'"
            warnings.warn(msg)


def get_backend():
    """Get backend of spartan.

    Returns
    -------
    str
        Backend name
    """
    return _BACKEND


def set_default_backend(backend_name):
    """Save default backend setting to config file.

    Parameters
    ----------
    backend_name : str, {'cpu', 'gpu'}
        Default backend name
    """
    default_dir = os.path.join(os.path.expanduser('~'), '.spartan')
    if not os.path.exists(default_dir):
        os.makedirs(default_dir)
    config_path = os.path.join(default_dir, 'config.json')
    if os.path.exists(config_path):
        configs = json.load(open(config_path, 'r'))
        configs['backend'] = backend_name.lower()
        json.dump(configs, open(config_path, 'w'))
    print(f'Setting the default backend to {backend_name}.')
    print(f'The config is saved at ~/.spartan/config.json.')
    print(f'Environment variable `SPARTAN_BACKEND` is prefered.')


def get_preferred_backend():
    """Get preferred backend from environment variable or config file.

    Returns
    -------
    str
        Backend name.
    """
    config_path = os.path.join(
        os.path.expanduser('~'), '.spartan', 'config.json')
    backend_name = 'cpu'
    if "SPARTAN_CONFIG" in os.environ:
        backend_name = os.getenv('SPARTAN_BACKEND')
    elif os.path.exists(config_path):
        with open(config_path, "r") as config_file:
            config_dict = json.load(config_file)
            backend_name = config_dict.get('backend', '').lower()

    backend_name = backend_name.lower()
    if backend_name in _backend_dict:
        return backend_name
    else:
        print(
            f"Spartan doesn't support backend {backend_name} yet. \
                Use cpu backend instead.")
        set_default_backend('cpu')
        return 'cpu'


load_backend(get_preferred_backend())



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py
@Desc    :   Import part and configuration part, including names exposed to spartan.
'''

# here put the import lib

from spartan.tensor import TensorData, TensorStream

from .ioutil import loadTensor, File, loadTensorStream, loadFile2Dict, loadHistogram
from .basicutil import set_trace, TimeMapper, StringMapper,\
    IntMapper, ScoreMapper, DenseIntMapper
from .drawutil import plot_graph, plot_timeseries, drawEigenPulse, plot, histogram_viz, clusters_viz, drawHexbin, drawRectbin
from .rect_histogram import RectHistogram
from .MinTree import MinTree

MODEL_PATH = 'spartan.model'

__all__ = [
    'loadTensor',
    'File',
    'loadTensorStream',
    'loadFile2Dict',
    'loadHistogram',
    'set_trace',
    'TimeMapper',
    'StringMapper',
    'IntMapper',
    'ScoreMapper',
    'DenseIntMapper',
    'plot',
    'plot_graph',
    'plot_timeseries',
    'drawEigenPulse',
    'histogram_viz',
    'clusters_viz',
    'drawHexbin',
    'drawRectbin',
    'RectHistogram',
    'MinTree',
]



from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext

ext_modules = [
    Extension("c_MDL",
        ["c_MDL.pyx"],
    )
]

setup(
    name="c_MDL",
    cmdclass = {"build_ext": build_ext},
    ext_modules = ext_modules
)



import builtins
import functools
import math

import torch

from .tensor import DTensor, STensor, _wrap_ret, _ensure_tensor


def _dispatch(dfunc, sfunc=None):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            is_sparse = builtins.any([isinstance(x, STensor) for x in args])
            is_dense = builtins.any([isinstance(x, DTensor) for x in args])
            if (is_sparse and is_dense):
                raise TypeError(
                    f"Parameters of `st.{func.__name__}` should be all STensor \
                    or all DTensor.")
            if is_sparse:
                if sfunc is None:
                    raise TypeError(f"`st.{func.__name__}` doesn't support \
                                    STensor parameters.")
                args = tuple(x._data if isinstance(x, STensor) else x
                             for x in args)
                return _ensure_tensor(sfunc(*args, **kwargs))
            else:
                args = tuple(x._data if isinstance(x, DTensor) else x
                             for x in args)
                return _ensure_tensor(dfunc(*args, **kwargs))
        return wrapper
    return decorator


# Type definition
short = torch.short
uint8 = torch.uint8
int8 = torch.int8
int16 = torch.int16
int32 = torch.int32
int64 = torch.int64
float16 = torch.float16
float32 = torch.float32
float64 = torch.float64
complex64 = torch.complex64
complex128 = torch.complex128


@_wrap_ret
def add(input_, other):
    """Wrapper of `torch.add`.

    Parameters
    ----------
    input_ : DTensor or STensor
        The first operand.
    other : DTensor or STensor
        The second operand.

    Returns
    -------
    DTensor or STensor
        Output tensor.
    """
    return torch.add(input_._data, other._data)


@_wrap_ret
def all(input_, axis=None, keepdims=False):
    """Wrapper of `torch.all`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False

    Returns
    -------
    DTensor or bool
        Output tensor.
    """
    if axis is None:
        return torch.all(input_._data)
    return torch.all(input_._data, dim=axis, keepdim=keepdims)


@_wrap_ret
def angle(input_, deg=False):
    """Wrapper of `torch.angle`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    deg : bool, optional
        If true, result is in degree format. Otherwise, return in radians. By
        default False
    """
    if deg:
        ret = torch.angle(input_) * 180 / math.pi
    else:
        ret = torch.angle(input_)
    return ret


@_wrap_ret
def any(input_, axis=None, keepdims=False):
    """Wrapper of `torch.any`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False

    Returns
    -------
    DTensor or bool:
    """
    if axis is None:
        return torch.any(input_._data)
    return torch.any(input_._data, dim=axis, keepdim=keepdims)


@_wrap_ret
def arange(start, stop, step, dtype=None):
    """Wrapper of `torch.arange`

    Parameters
    ----------
    start : number
        Start of the interval.
    stop : number
        End of the interval.
    step : number
        Spacing between spaces.
    dtype : data-type, optional
        Type of the return tensor, by default None

    Returns
    -------
    DTensor
        Constructed dense tensor.
    """
    return torch.arange(start, stop, step, dtype=dtype)


@_wrap_ret
def argmax(input_, axis=None):
    """Wrapper of `torch.argmax`

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    if axis is None:
        return torch.argmax(input_)
    else:
        return torch.argmax(input_, dim=axis)


@_wrap_ret
def argmin(input_, axis=None):
    """Wrapper of `torch.argmin`

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    if axis is None:
        return torch.argmin(input_)
    else:
        return torch.argmin(input_, dim=axis)


@_wrap_ret
def argsort(input_, axis=-1):
    """Wrapper of `torch.argsort`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    return torch.argsort(input_._data, dim=axis)


@_wrap_ret
def bincount(input_, weights=None, minlength=0):
    """Wrapper of `torch.bincount`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    weights : DTensor
        Weights.
    minlength : int, optional
        Minimum number of bins, by default 0
    """
    return torch.bincount(input_._data, weights, minlength)


@_wrap_ret
def bitwise_and(input_, other):
    """Wrapper of `torch.bitwise_and`

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.bitwise_and(input_._data, other._data)


@_wrap_ret
def bitwise_not(input_):
    """Wrapper of `torch.bitwise_not`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.bitwise_not(input_._data)


@_wrap_ret
def bitwise_or(input_, other):
    """Wrapper of `torch.bitwise_or`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.bitwise_or(input_._data, other._data)


@_wrap_ret
def bitwise_xor(input_, other):
    """Wrapper of `torch.bitwise_xor`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.bitwise_xor(input_._data, other._data)


def can_cast(from_, to):
    """Wrapper of `torch.can_cast`.

    Parameters
    ----------
    from_ : data-type
        Data type to cast from.
    to : data-type
        Data type to cast to.
    """
    return torch.can_cast(from_, to)


@_wrap_ret
def ceil(input_):
    """Wrapper of `torch.ceil`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.ceil(input_._data)


@_wrap_ret
def concatenate(inputs, axis=None):
    """Wrapper of `torch.cat`

    Parameters
    ----------
    inputs : sequence of DTensor
        Dense tensor to be concatenated.
    axis : int or None, optional
        Axis to operate on, by default None
    """
    return torch.cat((x._data for x in inputs), dim=axis)


@_wrap_ret
def conj(input_):
    """Wrapper of `torch.conj`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.conj(input_._data)


@_wrap_ret
def cos(input_):
    """Wrapper of `torch.cos`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.cos(input_._data)


@_wrap_ret
def cosh(input_):
    """Wrapper of `torch.cosh`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.cos(input_._data)


@_wrap_ret
def cumprod(input_, axis=None, dtype=None):
    """Wrapper of `torch.cumprod`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int, optional
        Axis to operate on, by default None
    dtype : data-type, optional
        Data type of output, by default None
    """
    return torch.cumprod(input_._data, dim=axis, dtype=dtype)


@_wrap_ret
def cumsum(input_, axis=None, dtype=None):
    """Wrapper of `torch.cumsum`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int, optional
        Axis to operate on, by default None
    dtype : data-type, optional
        Data type of output, by default None
    """
    return torch.cumsum(input_._data, axis, dtype)


@_wrap_ret
def diag(input_, k=0):
    """Wrapper of `torch.diag`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    k : int, optional
        Offset to main diagonal, by default 0
    """
    return torch.diag(input_._data, diagonal=k)


@_wrap_ret
def diagflat(input_, k=0):
    """Wrapper of `torch.diagflat`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    offset : int, optional
        Offset to main diagonal, by default 0
    """
    return torch.diagflat(input_._data, k)


@_wrap_ret
def diagonal(input_, offset=0, axis1=0, axis2=1):
    """Wrapper of `torch.diagonal`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    offset : int, optional
        Offset to the main diagonal, by default 0
    axis1 : int, optional
        The first axis of the diagonal, by default 0
    axis2 : int, optional
        The second axis of the diagonal, by default 1
    """
    return torch.diagonal(input_._data, offset=offset, dim1=axis1, dim2=axis2)


@_wrap_ret
def dot(input_, other):
    """Wrapper of `torch.dot`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    if input_._data.ndim == 1 and other._data.ndim == 1:
        return torch.dot(input_._data, other._data)
    return torch.matmul(input_._data, other._data)


@_wrap_ret
def empty(shape, dtype):
    """Wrapper of `torch.empty`.

    Parameters
    ----------
    shape : tuple of ints
        Shape of tensor.
    dtype : data-type
        Data type of tensor.
    """
    return torch.empty(shape, dtype=dtype)


@_wrap_ret
def empty_like(input_, dtype):
    """Wrapper of `torch.empty_like`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    dtype : data-type
        Data type of output.
    """
    return torch.empty_like(input_._data, dtype=dtype)


@_wrap_ret
def equal(input_, other):
    """Wrapper of `torch.equal`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.equal(input_._data, other._data)


@_wrap_ret
def exp(input_):
    """Wrapper of `torch.exp`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.exp(input_._data)


@_wrap_ret
def expm1(input_):
    """Wrapper of `torch.expm1`.

    Parameters
    ----------
    input_ : DTensor or STensor.
        Input tensor.
    """
    return torch.expm1(input_._data)


@_wrap_ret
def eye(n, m=None, dtype=None):
    """Wrapper of `torch.eye`.

    Parameters
    ----------
    n : int
        Number of rows.
    m : int, optional
        Number of cols, by default None (equal to `n`)
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    return torch.eye(n, m, dtype=dtype)


@_wrap_ret
def flip(input_, axis=None):
    """Wrapper of `torch.flip`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    if axis is None:
        axis = list(range(input_.data.ndim))
    return torch.flip(input_._data, axis)


@_wrap_ret
def floor(input_):
    """Wrapper of `torch.floor`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.floor(input_._data)


@_wrap_ret
def floor_divide(input_, other):
    """Wrapper of `torch.floor_divide`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.floor_divide(input_._data, other._data)


@_wrap_ret
def fmod(input_, other):
    """Wrapper of `torch.fmod`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.fmod(input_._data, other._data)


@_wrap_ret
def full(shape, value, dtype=None):
    """Wrapper of `torch.full`.

    Parameters
    ----------
    shape : tuple of ints
        Shape of output tensor.
    value : scalar
        Fill value of output tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    return torch.full(shape, value, dtype=dtype)


@_wrap_ret
def full_like(input_, value, dtype=None):
    """Wrapper of `numpy.full_like`.

    Parameters
    ----------
    input_ : DTensor
        The input tensor.
    value : scalar
        Fill value of output tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    return torch.full_like(input_._data, value, dtype=dtype)


@_wrap_ret
def imag(input_):
    """Wrapper of `torch.imag`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.imag(input_._data)


@_wrap_ret
def isfinite(input_):
    """Wrapper of `torch.isinfinite`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.isfinite(input_._data)


@_wrap_ret
def isinf(input_):
    """Wrapper of `torch.isinf`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.isinf(input_._data)


@_wrap_ret
def isnan(input_):
    """Wrapper of `torch.isnan`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.isnan(input_._data)


@_wrap_ret
def linspace(start, end, step, dtype=None):
    """Wrapper of `torch.linspace`.

    Parameters
    ----------
    start : DTensor
        Dense tensor of start points.
    end : DTensor
        Dense tensor of end points.
    step : int
        Number of sequence.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    return torch.linspace(start, end, step, dtype=dtype)


@_wrap_ret
def log(input_):
    """Wrapper of `torch.log`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.log(input_._data)


@_wrap_ret
def log10(input_):
    """Wrapper of `torch.log10`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.log10(input_._data)


@_wrap_ret
def log1p(input_):
    """Wrapper of `torch.log1p`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.log1p(input_._data)


@_wrap_ret
def log2(input_):
    """Wrapper of `torch.log2`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.log2(input_._data)


@_wrap_ret
def logical_and(input_, other):
    """Wrapper of `torch.logical_and`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.logical_and(input_._data, other._data)


@_wrap_ret
def logical_not(input_):
    """Wrapper of `torch.logical_not`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.logical_not(input_._data)


@_wrap_ret
def logical_or(input_, other):
    """Wrapper of `torch.logical_or`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.logical_or(input_._data, other._data)


@_wrap_ret
def logical_xor(input_, other):
    """Wrapper of `torch.logical_xor`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.logical_xor(input_._data, other._data)


@_wrap_ret
def logspace(start, stop, step, base=10, dtype=None):
    """Wrapper of `torch.logspace`.

    Parameters
    ----------
    start : DTensor
        Dense tensor of start points.
    stop : DTensor
        Dense tensor of end points.
    step : int
        Number of sequence.
    base : int, optional
        Base of the logarithm, by default 10
    dtype : data-type, optional
        Data type of output, by default None
    """
    return torch.logspace(start, stop, step, base=base, dtype=dtype)


@_wrap_ret
def matmul(input_, other):
    """Wrapper of `torch.matmul`.

    Parameters
    ----------
    input_ : DTensor or STensor
        The first operand.
    other : DTensor or STensor
        The second operand.
    """
    return torch.matmul(input_._data, other._data)


@_wrap_ret
def mean(input_, axis=None, keepdims=False):
    """Wrapper of `torch.mean`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    if axis is None:
        ret = torch.mean(input_._data)
    else:
        ret = torch.mean(input_._data, dim=axis, keepdim=keepdims)
    return ret


@_wrap_ret
def median(input_, axis=-1, keepdims=False):
    """Wrapper of `torch.median`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    if axis is None:
        ret = torch.median(input_._data)
    else:
        ret = torch.median(input_._data, dim=axis, keepdim=keepdims)
    return ret


@_wrap_ret
def meshgrid(*inputs):
    """Wrapper of `numpy.meshigrid`.
    """
    datas = [i._data for i in inputs]
    return tuple([d for d in torch.meshgrid(*datas)])


@_wrap_ret
def nonzero(input_):
    """Wrapper of `torch.nonzero`

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return tuple([d for d in torch.nonzero(input_._data, as_tuple=True)])


@_wrap_ret
def ones(shape, dtype=None):
    """Wrapper of `torch.ones`.

    Parameters
    ----------
    shape : tuple of ints
        Shape of output tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    return torch.ones(shape, dtype=dtype)


@_wrap_ret
def ones_like(input_, dtype=None):
    """Wrapper of `torch.ones_like`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    return torch.ones_like(input_._data, dtype=dtype)


@_wrap_ret
def prod(input_, axis=None, keepdims=False, dtype=None):
    """Wrapper of `torch.prod`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    dtype : data-type, optional
        Data type of output, by default None
    """
    if axis is None:
        return torch.prod(input_._data, dtype=dtype)
    return torch.prod(input_._data, dim=axis, keepdim=keepdims, dtype=dtype)


@_wrap_ret
def real(input_):
    """Wrapper of `torch.real`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.real(input_._data)


@_wrap_ret
def reciprocal(input_):
    """Wrapper of `torch.reciprocal`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.reciprocal(input_._data)


@_wrap_ret
def remainder(input_, other):
    """Wrapper of `torch.remainder`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.remainder(input_._data, other._data)


@_wrap_ret
def reshape(input_, shape):
    """Wrapper of `numpy.rehsape`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    shape : tuple of ints
        Shape of new tensor.
    """
    return torch.rehsape(input_._data, shape)


@_wrap_ret
def roll(input_, shift, axis=None):
    """Wrapper of `torch.roll`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    shift : int or tuple of ints
        Shift numbers of each axis or axes.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    return torch.roll(input_._data, shift, dims=axis)


@_wrap_ret
def rot90(input_, k=1, axes=(0, 1)):
    """Wrapper of `torch.rot90`

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    k : int, optional
        Number of rotation times, by default 1
    axes : tuple, optional
        The axes in which the input is rotated, by default (0, 1)
    """
    return torch.rot90(input_._data, k, dims=axes)


@_wrap_ret
def sign(input_):
    """Wrapper of `torch.sign`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.sign(input_._data)


@_wrap_ret
def sin(input_):
    """Wrapper of `torch.sin`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.sin(input_._data)


@_wrap_ret
def sinh(input_):
    """Wrapper of `torch.sinh`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.sinh(input_._data)


@_wrap_ret
def split(input_, indices_or_sections, axis=0):
    """Wrapper of `torch.split`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    indices_or_sections : int or DTensor
        If integer, specifing part of splits. Else specifing the positions
        where the splits take place.
    axis : int, optional
        The axis to operate on, by default 0
    """
    return torch.split(input_._data, indices_or_sections, axis)


@_wrap_ret
def sqrt(input_):
    """Wrapper of `torch.sqrt`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.sqrt(input_._data)


@_wrap_ret
def square(input_):
    """Wrapper of `torch.square`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.square(input_._data)


@_wrap_ret
def squeeze(input_, axis=None):
    """Wrapper of `torch.squeeze`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    return torch.squeeze(input_._data, dim=axis)


@_wrap_ret
def stack(inputs, axis=0):
    """Wrapper of `torch.stack`.

    Parameters
    ----------
    inputs : Sequence of DTensors.
        Dense tensors to be stacked, must have same shape.
    axis : int, optional
        Axis to operate on, by default 0
    """
    return torch.stack((x._data for x in inputs), dim=axis)


@_wrap_ret
def std(input_, axis=None, keepdims=False):
    """Wrapper of `torch.std`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    if axis is None:
        ret = torch.std(input_._data, unbiased=False)
    else:
        ret = torch.std(input_._data, dim=axis,
                        keepdim=keepdims, unbiased=False)
    return ret


@_wrap_ret
def sum(input_, axis=None, dtype=None, keepdims=False):
    """Wrapper of `numpy.sum`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    dtype : data-type, optional
        Data type of output, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    if axis is None:
        ret = torch.sum(input_._data, dtype=dtype)
    else:
        ret = torch.sum(input_._data, dim=axis, dtype=dtype, keepdim=keepdims)
    return ret


@_wrap_ret
def take(input_, indices):
    """Wrapper of `torch.take`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    indices : DTensor
        Indices of extracted values.
    """
    return torch.take(input_._data, indices)


@_wrap_ret
def tan(input_):
    """Wrapper of `torch.tan`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.tan(input_._data)


@_wrap_ret
def tanh(input_):
    """Wrapper of `torch.tanh`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.tanh(input_._data)


@_wrap_ret
def tensordot(input_, other, axes=2):
    """Wrapper of `numpy.tensordot`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor or STensor
        The second operand.
    axes : int or tuple of ints, optional
        Axis or axes along which the operation is performed, by default None
    """
    return torch.tensordot(input_._data, other._data, axes)


@_wrap_ret
def trace(input_):
    """Wrapper of `torch.trace`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    """
    return torch.trace(input_._data)


@_wrap_ret
def transpose(input_, axes=None):
    """Wrapper of `torch.transpose`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor
    axes : list of ints, optional
       Axes along which the operation is performed, by default None
    """
    if axes is None:
        axes = (0, 1)
    return torch.transpose(input_, axes[0], axes[1])


@_wrap_ret
def tril(input_, k=0):
    """Wrapper of `torch.tril`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    k : int, optional
        Offset to main diagonal, by default 0
    """
    return torch.tril(input_._data, k)


@_wrap_ret
def tril_indices(n, m=0, k=0):
    """Wrapper of `torch.tril_indices`.

    Parameters
    ----------
    n : int
        Number of row of output tensor.
    m : int, optional
        Number of column of output tensor, by default 0
    offset : int, optional
        Offset to main diagonal, by default 0
    """
    return torch.tril_indices(row=m, col=m, offset=k)


def triu(input_, k=0):
    """Wrapper of `torch.triu`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor
    k : int, optional
        Offset to main diagonal, by default 0
    """
    return torch.triu(input_._data, k)


@_wrap_ret
def triu_indices(n, m=0, k=0):
    """Wrapper of `torch.triu_indices`.

    Parameters
    ----------
    n : int
        Number of row of output tensor.
    m : int, optional
        Number of column of output tensor, by default 0
    offset : int, optional
        Offset to main diagonal, by default 0
    """
    ret = torch.triu_indices(row=m, col=m, offset=k)
    return tuple([index for index in ret])


@_wrap_ret
def true_divide(input_, other):
    """Wrapper of `torch.true_divide`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    return torch.true_divide(input_._data, other._data)


@_wrap_ret
def trunc(input_):
    """Wrapper of `torch.trunc`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    return torch.trunc(input_._data)


@_wrap_ret
def unique(input_, return_inverse=False, return_counts=False, axis=None):
    """Wrapper of `torch.unique`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    return_inverse : bool, optional
        If True, return the indices also, by default False
    return_counts : bool, optional
        If True, return the count also, by default False
    axis : int or None, optional
        The axis to operate on, by default None
    """
    return torch.unique(input_._data, return_inverse=return_inverse,
                        return_counts=return_counts, dim=axis)


@_wrap_ret
def var(input_, axis=None, keepdims=False):
    """Wrapper of `torch.var`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    if axis is None:
        ret = torch.var(input_)
    else:
        ret = torch.var(input_, dim=axis, keepdim=keepdims)
    return ret


@_wrap_ret
def where(condition, x, y):
    """Wrapper of `torch.where`.

    Parameters
    ----------
    condition : DTensor of bool
        Where True, yield x, otherwise yield y.
    x : DTensor
        The first tensor.
    y : DTensor
        The second tensor.
    """
    return torch.where(condition, x, y)


@_wrap_ret
def zeros(shape, dtype=None):
    """Wrapper of `torch.zeros`.

    Parameters
    ----------
    shape : tuple of ints
        Shape of output tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    return torch.zeros(shape, dtype=dtype)


@_wrap_ret
def zeros_like(input_, dtype=None):
    """Wrapper of `torch.zeros_like`.

    Parameters
    ----------
    input_ : DTensor
        Input tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    return torch.zeros_like(input_._data, dtype=dtype)



from spartan.util.basicutil import param_default

from .BeatGAN_CNN import BeatGAN_CNN
from .BeatGAN_RNN import BeatGAN_RNN


def __call__():
    return BeatGAN


class BeatGAN():
    def __init__(self, *args, **kwargs):
        raise ModuleNotFoundError("Please select a network type.")

    @classmethod
    def __create__(cls, *args, **kwargs):
        if kwargs['network'] == 'CNN':
            return BeatGAN_CNN(*args, **kwargs)
        elif kwargs['network'] == 'RNN':
            return BeatGAN_RNN(*args, **kwargs)
        else:
            return cls(*args, **kwargs)
